package be.iminds.ilabt.jfed.ui.cli;

import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetails;
import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetailsWriter;
import be.iminds.ilabt.jfed.log.ApiCallDetails;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.log.ResultListener;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AMv2Wrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AMv3Wrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticUserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.model.impl.BasicRspecFactory;
import be.iminds.ilabt.jfed.rspec.parser.RspecParseException;
import be.iminds.ilabt.jfed.rspec.parser.StaxRspecParser;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecMerger;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.ui.commandline.BaseCli;
import be.iminds.ilabt.jfed.util.*;
import ch.qos.logback.classic.LoggerContext;
import ch.qos.logback.classic.joran.JoranConfigurator;
import ch.qos.logback.core.joran.spi.JoranException;
import ch.qos.logback.core.util.StatusPrinter;
import com.google.inject.Injector;
import javanet.staxutils.IndentingXMLEventWriter;
import org.apache.commons.cli.*;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.FileConfiguration;
import org.apache.commons.configuration.PropertiesConfiguration;
import org.slf4j.LoggerFactory;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import java.io.*;
import java.security.PublicKey;
import java.util.*;
import java.util.stream.Collectors;

import static java.time.format.DateTimeFormatter.ISO_OFFSET_DATE_TIME;

/**
 * ExperimenterCli
 */
@SuppressWarnings("AccessStaticViaInstance")
public class ExperimenterCli {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(ExperimenterCli.class);

    private static final String KEY_EXPIRATION_TIME = "expiration-time";
    private static final String KEY_REQUEST_RSPEC = "request-rspec";

    private static final String CHOOSE_PROJECT_AUTOMATICALLY_OPTION = "CHOOSE_AUTOMATICALLY";

    public static void commandHelp(final PrintStream err, Options options, String command) {
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp(200, "jfed-experimenter-cli <command> [command_options ... ]", "Available commands: create,delete,status,renew,userinfo\n" + command + " options:", options,
                "\n\nTemplate Context Properties File:\n  username = <username>\n" +
                        "  passwordFilename = <filename of file containing password>\n" +
                        "  pemKeyAndCertFilename = <filename of file containing user certificate and private key in PEM format>\n" +
                        "  userAuthorityUrn = <URN of test user's authority>\n"
        );
    }

    public static void showVersion(PrintStream out) {
        out.println("jFed Experimenter GUI -- http://jfed.iminds.be/");
        out.println("Send bugreports to: jfed-bugreports@atlantis.ugent.be");

        final String VERSION_PROPERTIES_FILE = "/version.properties";
        final String PROP_VERSION = "version";
        final String PROP_GIT_COMMIT = "build.git_commit";
        final String PROP_GIT_BRANCH = "build.git_branch";
        final String PROP_BUILD_JENKINS = "build.jenkins_build";

        Properties prop = new Properties();
        //now load the version-properties file
        try {
            prop.load(ExperimenterCli.class.getResourceAsStream(VERSION_PROPERTIES_FILE));
        } catch (IOException e) {
            throw new RuntimeException("Something went wrong while retrieving the version info: " + e.getMessage(), e);
        }

        StringBuilder version = new StringBuilder();
        version.append(prop.getProperty(PROP_VERSION, "Unknown"));
        final String jenkinsBuildNr = prop.getProperty(PROP_BUILD_JENKINS);
        if (jenkinsBuildNr != null && !jenkinsBuildNr.isEmpty())
            version.append("  - build #").append(jenkinsBuildNr);

        final String gitCommit = prop.getProperty(PROP_GIT_COMMIT);
        if (gitCommit != null && !gitCommit.isEmpty())
            version.append("  - git commit #").append(gitCommit);

        final String gitBranch = prop.getProperty(PROP_GIT_BRANCH);
        if (gitBranch != null && !gitBranch.isEmpty())
            version.append(" on ").append(gitBranch);
        out.println("Version: " + version.toString());

        out.println("Environment: "
                + System.getProperty("os.name") + " " + System.getProperty("os.version") + " " + System.getProperty("os.arch")
                + " - Java " + System.getProperty("java.version") + " (" + System.getProperty("java.vendor") + ")");
    }

    public static void generalHelp(final PrintStream err) {
        err.println("Syntax: jfed-experimenter-cli <command> [command_options ... ]\n" +
                "Available Commands: create,delete,status,renew\n");
    }

    public static void main(String[] args) throws IOException, ClassNotFoundException {
        int retval = main_helper(args, System.out, System.err, System.in, new BaseCli(), true);
        System.exit(retval);
    }

    public static int main_helper(String[] args,
                                  final PrintStream out, final PrintStream err, final InputStream in,
                                  BaseCli baseCli,
                                  boolean temporarilyRedirectSystemStreams) throws IOException, ClassNotFoundException {
        final PrintStream orig_out = System.out;
        final PrintStream orig_err = System.err;
        final InputStream orig_in = System.in;
        if (temporarilyRedirectSystemStreams) {
            System.setOut(out);
            System.setErr(err);
            System.setIn(in);
        }

        int retval = -1;
        try {
            retval = main_internal(args, out, err, in, baseCli);
        } finally {
            if (temporarilyRedirectSystemStreams) {
                System.setOut(orig_out);
                System.setErr(orig_err);
                System.setIn(orig_in);
            }
        }
        return retval;
    }

    public static int main_internal(String[] args,
                                    final PrintStream out, final PrintStream err, final InputStream in,
                                    BaseCli baseCli) throws IOException, ClassNotFoundException {
        String command;
        boolean isCreate = false;
        boolean isRenew = false;
        boolean isDelete = false;
        boolean isStatus = false;
        boolean isListUserInfo = false;
        LinkedList<String> argList = new LinkedList<>(Arrays.asList(args));
        if (argList.isEmpty()) {
            generalHelp(err);
            return 3;
        } else {
            if (argList.get(0).equals("version") || argList.get(0).equals("--version") || argList.get(0).equals("-version") || argList.get(0).equals("-v")) {
                showVersion(out);
                return 0;
            } else {
                command = argList.removeFirst();
                if (command.equalsIgnoreCase("create")) isCreate = true;
                if (command.equalsIgnoreCase("createsliver")) isCreate = true;
                if (command.equalsIgnoreCase("createslivers")) isCreate = true;
                if (command.equalsIgnoreCase("renew")) isRenew = true;
                if (command.equalsIgnoreCase("delete")) isDelete = true;
                if (command.equalsIgnoreCase("status")) isStatus = true;
                if (command.equalsIgnoreCase("userinfo")) isListUserInfo = true;

                if (!isCreate && !isRenew && !isStatus && !isDelete && !isListUserInfo) {
                    err.println("ERROR: unknown command \"" + command + "\"");
                    generalHelp(err);
                    return 3;
                }
            }
        }

        CommandLineParser parser = new BasicParser();

        boolean requiresSlice = isCreate || isDelete || isRenew || isStatus;

        // create the Options
        Options options = new Options();
//        options.addOption(OptionBuilder.withLongOpt("context-file")
//                .withDescription("The properties file containing context details (only the login info in the context file is used by this tool)")
//                .hasArg()
//                .withArgName("CONTEXT PROPERTIES FILE")
//                .create('c'));
//
//        options.addOption(OptionBuilder.withLongOpt("cert-and-key-file")
//                .withDescription("The file containing the user certificate and private key (both in PEM format)")
//                .hasArg()
//                .withArgName("PEM FILE")
//                .create('p'));
//        options.addOption(OptionBuilder.withLongOpt("private-key-password")
//                .withDescription("The password of the private key. Only used if private key is password protected. Default: interactively ask password when needed")
//                .hasArg()
//                .withArgName("CLEARTEXT PASSWORD")
//                .create('P'));
//        //TODO add options to specify username and user authority if not in certificate
        BaseCli.addUserConfig(options, "The properties file containing context details (only the login info in the context file is used by this tool)", false);

        BaseCli.addAuthoritiesConfig(options);
        if (requiresSlice)
            options.addOption(OptionBuilder.withLongOpt("slice")
                    .withDescription("The URN or name of the slice to use. (auto detected)")
                    .hasArg()
                    .withArgName("SLICE URN OR NAME")
                    .isRequired()
                    .create('s'));
        options.addOption(OptionBuilder.withLongOpt("speaks-for")
                .withDescription("The speaksFor certificate (in a file), enabling you make calls on behalf of another user.")
                .hasArg()
                .withArgName("CERTIFICATE FILENAME")
                .create());
        if (requiresSlice)
            options.addOption(OptionBuilder.withLongOpt("project-name")
                    .withDescription("The name of the project (= sub authority) of the slice. This is an optional argument (however some authorities might require a project!). " +
                            "You can also use \"" + CHOOSE_PROJECT_AUTOMATICALLY_OPTION + "\" as project name, in that case, the last (determined by your SA) project you are a member of will be used.")
                    .hasArg()
                    .withArgName("PROJECT NAME")
                    .create('S'));
        if (isCreate || isRenew) {
            options.addOption(OptionBuilder.withLongOpt("expiration-date")
                    .withDescription("The date and time at which this slice expires. In RFC3339 format. This is an optional argument. default: 2 hours in the future")
                    .hasArg()
                    .withArgName("RFC3339 DATE")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("expiration-hours")
                    .withDescription("The number of hours after which this slice expires. This is an optional argument. default: 2 hours in the future")
                    .hasArg()
                    .withArgName("INTEGER")
                    .create('e'));
        }
        if (isListUserInfo)
            options.addOption(OptionBuilder.withLongOpt("output-format")
                    .withDescription("The output format to use (for user information). Choices: \"text\" or \"json\". Default:json")
                    .hasArg()
                    .withArgName("FORMAT")
                    .create());
        if (isCreate || isStatus || isRenew)
            options.addOption(OptionBuilder.withLongOpt("rspec")
                    .withDescription("The rspec file to use for creating a sliver")
                    .hasArg()
                    .withArgName("RSPEC XML FILE")
                    .isRequired()
                    .create('r'));
        if (isDelete)
            options.addOption(OptionBuilder.withLongOpt("rspec")
                    .withDescription("The rspec file to use for creating a sliver (not mandatory for delete, but might not work without)")
                    .hasArg()
                    .withArgName("RSPEC XML FILE")
                    .create('r'));
        if (isCreate) {
            options.addOption(OptionBuilder.withLongOpt("create-slice")
                    .withDescription("Create the slice if it doesn't exist. (if it exists, this option is ignored)")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("rewrite-rspec")
                    .withDescription("Parse the provided RSpec, and reconstruct it again, before sending it to the server. This can help prevent some server sides bug caused by valid but atypical RSpecs.")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("ssh-keys")
                    .withDescription("Specify which ssh keys to add. The argument is a (comma separated) list of options.\n" +
                            "Available options: usercert,userkeys,rspec\n" +
                            "   usercert: add the user making the calls, with the ssh key from the certificate in the login PEM file.\n" +
                            "   userkeys: add the user making the calls, with the ssh keys that are stored on the user authority MA server. (usercert and userkeys can be combined to add all keys)\n" +
                            "   rspec: add the users and keys specified in the RSpec itself.\n" +
                            "This option is optional, default: usercert,rspec")
                    .hasArg()
                    .withArgName("OPTION LIST")
                    .create('k'));
            options.addOption(OptionBuilder.withLongOpt("stitching")
                    .withDescription("If the RSpec requires stitching, allow it. By default, RSpecs that require stitching cause an abort")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("manifest")
                    .withDescription("The file in which the manifest must be stored. (default: manifest-<SLICENAME>.rspec)")
                    .hasArg()
                    .withArgName("FILE")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("call-log")
                    .withDescription("A file into which all call details will be stored")
                    .hasArg()
                    .withArgName("FILE")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("slice-recover-info")
                    .withDescription("A file into which slice recover info will be stored. This can be used by the jFed GUI (if copied to the correct directory).")
                    .hasArg()
                    .withArgName("FILE")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("nowait")
                    .withDescription("Do not wait until sliver is ready (default: wait until ready)")
                    .create());
        }
        options.addOption(OptionBuilder.withLongOpt("print-calls")
                .withDescription("Print all calls to stdout")
                .create());
        options.addOption(OptionBuilder.withLongOpt("am-api")
                .withDescription("The AM Api version to use (\"2\" and \"3\" supported, default: choose automatically)")
                .hasArg()
                .withArgName("AM API VERSION")
                .create("am"));
        options.addOption(OptionBuilder.withLongOpt("fake")
                .withDescription("Do everything, expect actually making the slice and sliver calls. Useful for debugging syntax. (will make calls relating to retrieving user data)")
                .create());
        options.addOption(OptionBuilder.withLongOpt("fake-sliver")
                .withDescription("Do everything, expect actually making the sliver calls. (will make calls relating to retrieving user data and creating or retrieving slice)")
                .create());
        options.addOption("q", "quiet", false, "less output");
        options.addOption("d", "debug", false, "extra debugging output");
        options.addOption("l", "logging", false, "activate logback logging output");
        options.addOption("v", "version", false, "show version and exit");

        CommandLine line;
        try {
            // parse the command line arguments
            String[] leftoverArgs = new String[argList.size()];
            int i = 0;
            for (String arg : argList)
                leftoverArgs[i++] = arg;
            line = parser.parse(options, leftoverArgs);
        } catch (ParseException exp) {
            err.println("Command line argument Syntax error: " + exp.getMessage());
            commandHelp(err, options, command);
            return 3;
        }

        if (line.hasOption("version")) {
            showVersion(out);
            return 0;
        }

        final String sliceIdArg = line.getOptionValue("slice");
        String projectNameArg = line.getOptionValue("project-name");
        final String sliceExpirationDateArg = line.getOptionValue("expiration-date");
        final String sliceExpirationHoursArg = line.getOptionValue("expiration-hours");
        final String rspecArg = line.getOptionValue("rspec");
        final String outputFormatArg = line.getOptionValue("output-format") == null ? "json" : line.getOptionValue("output-format");
        final String amApiVersionArg = line.getOptionValue("am-api");
//        final String testContextFilenameArg = line.getOptionValue("context-file");
//        final String pemFilenameArg = line.getOptionValue("cert-and-key-file");
//        final String privateKeyPassArg = line.getOptionValue("private-key-password");
//        final String testAuthoritiesFilenameArg = line.getOptionValue("authorities-file");
        final String sshKeysArg = line.getOptionValue("ssh-keys");
        String manifestArg = line.getOptionValue("manifest");
        String callLogArg = line.getOptionValue("call-log");
        String sliceRecoverInfoArg = line.getOptionValue("slice-recover-info");
        final boolean createSliceArg = line.hasOption("create-slice");
        final boolean rewriteRspecArg = line.hasOption("rewrite-rspec");
        final boolean stitchingArg = line.hasOption("stitching");
        final boolean debug = line.hasOption("debug");
        final boolean printCalls = line.hasOption("print-calls");
        final boolean fake = line.hasOption("fake");
        final boolean logbackLogging = line.hasOption("logging");
        final boolean fakeSliver = line.hasOption("fake-sliver");
        final boolean waitUntilReady = !line.hasOption("nowait");
        final boolean silent = !debug && line.hasOption("quiet");
        final String speaksForArg = line.getOptionValue("speaks-for");

        if (debug) {
            out.println("Showing version because debug is enabled:");
            showVersion(out);
        }

        if (logbackLogging) {
            LoggerContext loggerContext = (LoggerContext) LoggerFactory.getILoggerFactory();
            loggerContext.reset();
            JoranConfigurator configurator = new JoranConfigurator();
            InputStream configStream = ExperimenterCli.class.getResourceAsStream("/logback_debug.xml");
            configurator.setContext(loggerContext);
            try {
                configurator.doConfigure(configStream); // loads logback file
            } catch (JoranException e) {
                LOG.error("Failed to configure logback", e);
                err.println("Failed to configure logback! "+e.getMessage());
                e.printStackTrace();
            }
            StatusPrinter.printInCaseOfErrorsOrWarnings(loggerContext);

            if (configStream != null)
                configStream.close();
        }

        if (requiresSlice) {
            //should not be required to check this here, as it is a required option, but let's do it anyway to prevent any issues.
            if (sliceIdArg == null) {
                err.println("The slice option is mandatory");
                commandHelp(err, options, command);
                return 3;
            }
        }

        boolean isJson = false;
        boolean isTextOutput = false;
        if (isListUserInfo) {
            boolean knownFormat = false;
            if (outputFormatArg.trim().equalsIgnoreCase("json")) {
                isJson = true;
                knownFormat = true;
            }
            if (outputFormatArg.trim().equalsIgnoreCase("text")) {
                isTextOutput = true;
                knownFormat = true;
            }
            if (!knownFormat) {
                err.println("Output format \"" + outputFormatArg + "\" is unknown");
                commandHelp(err, options, command);
                return 3;
            }
        }

        boolean sshKeysUserCert = false;
        boolean sshKeysUserServerKeys = false;
        boolean sshKeysRspec = false;

        ExtraInfoCallback.setBonfireLoginCallback(new ExtraInfoCallback.BonfireLoginCallback() {
            private String cachedGroup = null;

            @Override
            public String getBonfireGroups() {
                LOG.info("Bonfire groups callback called.");
                if (cachedGroup == null) {
                    out.println("\n");
                    out.println("==============================================================");
                    cachedGroup = IOUtils.askCommandLineInput("Please enter \"Bonfire Group for experiment\":", in);
                    out.println("==============================================================\n");
                }
                return cachedGroup;
            }
        });
        ExtraInfoCallback.setAnyLoginCallback(new ExtraInfoCallback.AnyLoginCallback() {
            private JFedPasswordManager.LoginInfo cachedLogin = null;

            @Override
            public JFedPasswordManager.LoginInfo getLogin(String serviceName, String friendlyName) {
                LOG.info("Login callback called.");
                if (cachedLogin == null) {
                    out.println("\n");
                    out.println("==============================================================");
                    String username = IOUtils.askCommandLineInput("Please enter "+friendlyName+" username:", in);
                    String password = new String(IOUtils.askCommandLinePassword("Please enter "+friendlyName+" password:", in));
                    cachedLogin = new JFedPasswordManager.LoginInfo(serviceName, username, password);
                    out.println("==============================================================\n");
                }
                return cachedLogin;
            }
        });

        if (debug)
            for (Option o : line.getOptions())
                out.println("ARG: " + o.getOpt() + " " + o.getLongOpt() + " " + o.getValue() + " " + o.getValuesList() + " " + o.getValuesList().size());

        if (sshKeysArg != null) {
            String[] opts = sshKeysArg.split(",");
            for (String opt : opts) {
                boolean validOpt = false;

                if (opt.equalsIgnoreCase("usercert")) {
                    sshKeysUserCert = true;
                    validOpt = true;
                }
                if (opt.equalsIgnoreCase("userkeys")) {
                    sshKeysUserServerKeys = true;
                    validOpt = true;
                }
                if (opt.equalsIgnoreCase("rspec")) {
                    sshKeysRspec = true;
                    validOpt = true;
                }

                if (!validOpt) {
                    err.println("The ssh-keys option you have specified (" + opt + ") is not valid");
                    commandHelp(err, options, command);
                    return 3;
                }
            }
        } else {
            //default: usercert,rspec
            sshKeysUserCert = true;
            sshKeysUserServerKeys = false;
            sshKeysRspec = true;
        }

        if (requiresSlice && sliceIdArg.startsWith("urn:")) {
            GeniUrn sliceUrn = GeniUrn.parse(sliceIdArg);
            if (sliceUrn == null) {
                err.println("Error parsing provided slice URN: " + sliceIdArg);
                return 2;
            }
            String sliceUrnprojectName = sliceUrn.getEncodedSubAuthName();
            if (((projectNameArg != null) != (sliceUrnprojectName != null)) ||
                    (projectNameArg != null && sliceUrnprojectName != null &&
                            !projectNameArg.equals(sliceUrnprojectName))) {
                if (projectNameArg != null) {
                    if (projectNameArg.equals(CHOOSE_PROJECT_AUTOMATICALLY_OPTION)) {
                        projectNameArg = sliceUrnprojectName;
                    } else {
                        err.println("The project-name you have specified (" + projectNameArg + ") does not match the sub authority in the slice URN you have specified (" + sliceIdArg + ")");
                        return 3;
                    }
                } else
                    projectNameArg = sliceUrnprojectName;
            }
        }
        final String sliceName = !requiresSlice ? null :
                (sliceIdArg.startsWith("urn:") ? GeniUrn.parse(sliceIdArg).getEncodedResourceName() : sliceIdArg);

        if (manifestArg == null)
            manifestArg = "manifest-" + sliceName + ".rspec";

        if (sliceExpirationDateArg != null && sliceExpirationHoursArg != null) {
            err.println("You may not specify both expiration-date and expiration-hours. You need to specify only one.");
            return 3;
        }

        final Date requestedExpirationDate;
        if (sliceExpirationDateArg == null) {
            if (sliceExpirationHoursArg == null)
                requestedExpirationDate = new Date(System.currentTimeMillis() + (2 * 60 * 60 * 1000l)); //2 hours
            else {
                try {
                    int hours = Integer.parseInt(sliceExpirationHoursArg);
                    requestedExpirationDate = new Date(System.currentTimeMillis() + (hours * 60 * 60 * 1000l));
                } catch (NumberFormatException e) {
                    err.println("Specified slice expiration hours (\"" + sliceExpirationDateArg + "\") is not a valid number.");
                    return 3;
                }
            }
        } else {
            try {
                requestedExpirationDate = RFC3339Util.rfc3339StringToDate(sliceExpirationDateArg);
            } catch (java.text.ParseException e) {
                err.println("Specified slice expiration date (\"" + sliceExpirationDateArg + "\") is not in RFC3339 format. The current date in RFC3339 format is: " + RFC3339Util.dateToRFC3339String(new Date()));
                return 3;
            }
        }


        baseCli.initAuthoritiesModule(line);
        try {
            baseCli.initUserModule(line, out, err, in);
        } catch (BaseCli.CliArgumentException e) {
            LOG.debug("Got CliArgumentException", e);
            commandHelp(err, options, command);
            return 3;
        }
        final Injector injector = baseCli.getInjector(err);
        final AuthorityListModel authorityListModel = injector.getInstance(AuthorityListModel.class);
        final AuthorityFinder authorityFinder = injector.getInstance(AuthorityFinder.class);
        final GeniUser user = injector.getInstance(GeniUser.class);
        final JFedPreferences jFedPreferences = injector.getInstance(JFedPreferences.class);
        final Logger logger = injector.getInstance(Logger.class);

        final CallLogOutput callLogOutput = new CallLogOutput(err);
        if (callLogArg != null) {
            callLogOutput.fos = new FileOutputStream(callLogArg);

            try {
                XMLOutputFactory xof = XMLOutputFactory.newInstance();
                callLogOutput.xtw = new IndentingXMLEventWriter(xof.createXMLEventWriter(callLogOutput.fos, "UTF-8"));
                callLogOutput.eventFactory = XMLEventFactory.newInstance();

                callLogOutput.xtw.add(callLogOutput.eventFactory.createStartDocument());
                callLogOutput.xtw.add(callLogOutput.eventFactory.createStartElement(new QName("calls"), null, null));
            } catch (XMLStreamException e) {
                callLogOutput.onFailure();
                LOG.error("Exception generating XML for call output", e);
                err.println("Exception generating XML for call output: " + e.getMessage());
            }

            logger.addResultListener(result -> {
                if (callLogOutput.isDisabled())
                    return;

                SerializableApiCallDetails serializableApiCallDetails = new SerializableApiCallDetails(result);

                try {
                    if (result.getAuthority() != null)
                        callLogOutput.xtw.add(callLogOutput.eventFactory.createComment("" + result.getGeniMethodName() + " on " + result.getAuthority().getUrnString() + ""));

                    SerializableApiCallDetailsWriter.writeApiCallDetails(serializableApiCallDetails, callLogOutput.xtw, callLogOutput.eventFactory);

                    //add 2 newlines
                    callLogOutput.xtw.add(callLogOutput.eventFactory.createCharacters(System.getProperty("line.separator") + System.getProperty("line.separator")));

                    callLogOutput.xtw.flush();
                    callLogOutput.fos.flush();
                } catch (IOException e) {
                    err.println("Error flushing: " + e.getMessage());
                    e.printStackTrace();
                } catch (XMLStreamException e) {
                    err.println("Error writing call to stream: " + e.getMessage());
                    e.printStackTrace();

                    callLogOutput.onFailure();
                }
            });
        } else {
            callLogOutput.fos = null;
            callLogOutput.xtw = null;
            callLogOutput.eventFactory = null;
        }
        if (printCalls) {
            PrintStream printCallsOut = out;
            final String printCallsPrefix = "CALL: ";

            logger.addResultListener(new ResultListener() {
                @Override
                public void onResult(ApiCallDetails result) {
                    printCallsOut.println();
                    printCallsOut.println(TextUtil.indent(printCallsPrefix.length()+1, printCallsPrefix + "" + result.getGeniMethodName() +
                            " on " + result.getAuthority().getUrnString() + "" +
                            " at "+result.getConnectionConfig().getServerUrlString()));

                    String req = result.getXmlRpcRequestJsonString();
                    String res = result.getXmlRpcResponseJsonString();
                    if (req != null) printCallsOut.println(printCallsPrefix + "Request: " + req);
                    if (res != null) printCallsOut.println(printCallsPrefix + "Reply: " + res);

                    printCallsOut.println();
                    printCallsOut.println();
                    printCallsOut.flush();
                }
            });
        }
        try { //used for finally callLogOutput.close()
            if (amApiVersionArg != null && !amApiVersionArg.equals("2") && !amApiVersionArg.equals("3")) {
                err.println("Unsupported AM api version: \"" + amApiVersionArg + "\"");
                commandHelp(err, options, command);
                return 3;
            }

            String rspecString;
            StringRspec rspec = null;
            List<String> rspecComponentManagers = null;
            if (rspecArg != null) {
                assert rspecArg != null;
                try {
                    rspecString = IOUtils.fileToString(rspecArg);
                } catch (IOException e) {
                    err.println("The RSpec file \"" + rspecArg + "\" was not found.");
                    return 3;
                }
                rspec = new StringRspec(rspecString);
                rspecComponentManagers = rspec.getAllComponentManagerUrns();
                if (rspecComponentManagers.size() != 1 || debug) {
                    if (!silent)
                        out.println("Note: This RSpec has " + rspecComponentManagers.size() + " component managers: " + rspecComponentManagers);
                }
                if (rspecComponentManagers.size() == 0) {
                    if (isCreate) {
                        err.println("ERROR: The provided RSpec has no component managers. Something went wrong during parsing, or this is an empty RSpec. Either way, 'createSliver' cannot continue.");
                        return 3;
                    }
                    else
                        err.println("WARNING: The provided RSpec has no component managers. Something went wrong during parsing, or this is an empty RSpec. Either way, few commands can now meaningfully continue. Will try anyway.");
                }
            } else {
                //no RSpec provided, get rspecComponentManagers from SLIVER_INFO (or equivalent)
                //this is done below, after user info is retrieved
            }

            if (stitchingArg && rspec != null) {
                if (!rspec.isStitching(authorityListModel, authorityFinder)) {
                    err.println("The --stitching option was set, but the rspec does not specify any stitched links: Aborted CreateSliver.");
                    return 3;
                }
            }

            if (!BaseCli.checkUser(user, line, out, err, in))
                return 3;

            final SfaAuthority userAuth = user.getUserAuthority();
            final GeniUserProvider userProvider = injector.getInstance(GeniUserProvider.class);

            if (!silent) {
            }
            if (debug) {
                out.println("User " + user.getUserUrnString() + " will be used.");
                out.println("User Authority: " + user.getUserAuthority().getName());
                out.println("User Authority URN: " + user.getUserAuthority().getUrnString());
            }

            boolean usesSpeaksFor = speaksForArg != null;
            GeniUrn speakingForUser = null;
            AnyCredential speaksForCredential = null;
            if (usesSpeaksFor) {
                String speaksForCredentialString;
                try {
                    speaksForCredentialString = IOUtils.fileToString(speaksForArg);
                } catch (IOException e) {
                    LOG.error("Error reading speaksFor credential from file \"" + speaksForArg + "\"", e);
                    err.println("Error reading speaksFor credential from file \"" + speaksForArg + "\": " + e.getMessage());
                    return 3;
                }
                try {
                    speaksForCredential = AnyCredential.createAutoDetect("speaksFor", speaksForCredentialString);
                } catch (CredentialException e) {
                    LOG.error("Error parsing speaksFor credential from file \"" + speaksForArg + "\"", e);
                    err.println("Error parsing speaksFor credential from file \"" + speaksForArg + "\": " + e.getMessage());
                    return 3;
                }
                if (speaksForCredential == null) {
                    LOG.error("Error parsing speaksFor credential from file \"" + speaksForArg + "\"");
                    err.println("Error parsing speaksFor credential from file \"" + speaksForArg + "\"");
                    return 3;
                }
                if (!speaksForCredential.isSpeaksFor()) {
                    LOG.warn("WARNING: parsed speaksFor credential from file \"" + speaksForArg + "\" is NOT a speaksFor credential");
                    err.println("WARNING parsed speaksFor credential from file \"" + speaksForArg + "\" is NOT a speaksFor credential. (Will continue anyway)");
                }
                try {
                    if (speaksForCredential instanceof AbacCredential) {
                        AbacCredential c = (AbacCredential) speaksForCredential;
                        if (c.getSpokenForUrn() != null)
                            speakingForUser = c.getSpokenForUrn();
                    }
                    if (speaksForCredential instanceof SfaCredential) {
                        SfaCredential c = (SfaCredential) speaksForCredential;
                        if (c.getTargetUrn() != null)
                            speakingForUser = new GeniUrn(c.getTargetUrn());
                        else {
                            if (c.getTargetGid() != null) {
                                Gid gid = new Gid(c.getTargetGid());
                                if (gid.getSubjectUrn() != null) {
                                    speakingForUser = new GeniUrn(gid.getSubjectUrn());
                                }
                            }
                        }
                    }
                } catch (GeniUrn.GeniUrnParseException e) {
                    LOG.error("Error parsing spoken for user URN from speaksFor credential from file \"" + speaksForArg + "\"", e);
                    err.println("Error parsing spoken for user URN from speaksFor credential from file \"" + speaksForArg + "\": " + e.getMessage());
                    return 3;
                }
//                if (!speaksForCredential.validate()) {
//                    LOG.warn("WARNING: Provided speaksfor credential does not validate");
//                    err.println("WARNING: Provided speaksfor credential does not validate");
//                }

                boolean check;
                try {
                    check = speaksForCredential.check(authorityListModel);
                } catch (CredentialException e) {
                    LOG.warn("Error trying to check credential signature: ", e);
                    check = false;
                }
                if (!check) {
                    LOG.warn("WARNING: Provided speaksfor credential signature check failed. (frequently caused by: you modified the credential)");
                    err.println("WARNING: Provided speaksfor credential signature check failed. (frequently caused by: you modified the credential)");
                }

                if (speakingForUser == null) {
                    LOG.warn("ERROR: could not determine the spoken for user in the speaksFor credential from file \"" + speaksForArg + "\". Cannot continue.");
                    err.println("ERROR: could not determine the spoken for user in the speaksFor credential from file \"" + speaksForArg + "\". Cannot continue.");
                    return 3;
                }
                if (!silent)
                    out.println("Using speaksFor credential for user \"" + speakingForUser + "\"");
            }

            JFedConnectionProvider conPool = injector.getInstance(JFedConnectionProvider.class);
            AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory
                    automaticUserAndSliceApiWrapperFactory =
                    injector.getInstance(AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory.class);
            AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory
                    automaticAggregateManagerWrapperFactory =
                    injector.getInstance(AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory.class);

            final UserAndSliceApiWrapper sa = automaticUserAndSliceApiWrapperFactory.create();
            GeniUrn targetUserUrn = user.getUserUrn();
            if (usesSpeaksFor) {
                sa.setSpeaksFor(speaksForCredential, speakingForUser);
                targetUserUrn = speakingForUser;
            }
            String targetUserUrnString = targetUserUrn.toString();

            List<AnyCredential> userCreds;
            try {
                userCreds = sa.getLoggedinUserCredentials(logger);
            } catch (JFedException e) {
                return callFailed("Error retrieving user credential", e, err, callLogOutput);
            }
            if (debug)
                out.println(userCreds.size() + " user credentials retrieved.");
            List<GeniUrn> existingSlices;
            try {
                existingSlices = sa.getSlicesForUser(logger, targetUserUrn);
            } catch (JFedException e) {
                return callFailed("Error retrieving slices for user", e, err, callLogOutput);
            }
            if (debug)
                out.println("User " + targetUserUrnString + " has the following slices: " + existingSlices);

            if (isListUserInfo) {
                List<String> projectNames;
                try {
                    projectNames = sa.getSubAuthorityNames(logger, targetUserUrn);
                } catch (JFedException e) {
                    return callFailed("Error retrieving projects for user", e, err, callLogOutput);
                }
                if (isJson) {
                    Hashtable outerUserHt = new Hashtable();
                    Hashtable userHt = new Hashtable();
                    userHt.put("urn", targetUserUrnString);
                    userHt.put("slices", new Vector(existingSlices));
                    userHt.put("projects", new Vector(projectNames));
                    outerUserHt.put("User Info", userHt);
//                    out.println(JSonHelper.jsonObjectToString(outerUserHt));
                    out.println(JSonHelper.jsonObjectToString(userHt));
                }
                if (isTextOutput) {
                    out.println("User Info: ");
                    out.println("   Urn: "+targetUserUrnString);
                    out.println("   Projects: "+projectNames.stream().collect(Collectors.joining(",")).toString());
                    out.println("   Slices: "+existingSlices.stream().map(u -> u.toString()).collect(Collectors.joining(",")).toString());
                }

                return 0;
            }

            List<String> projectNames = null;
            if (projectNameArg != null && projectNameArg.equals(CHOOSE_PROJECT_AUTOMATICALLY_OPTION)) {
                try {
                    projectNames = sa.getSubAuthorityNames(logger, targetUserUrn);

                    if (debug)
                        out.println("You are a member of the following projects: " + projectNames);

                    if (projectNames.isEmpty()) {
                        err.println("WARNING: " + CHOOSE_PROJECT_AUTOMATICALLY_OPTION + " was set as project, but you do not seem to be a member of any project! Will use fallback: continue without project.");

                        //fallback: no project
                        projectNameArg = null;
                    } else {
                        //select last project
                        projectNameArg = projectNames.get(projectNames.size() - 1);
                        if (debug || !silent)
                            out.println("Automatically selected project \"" + projectNameArg + "\"");
                    }
                } catch (JFedException e) {
                    //fallback: no project
                    projectNameArg = null;

                    String message =
                            "WARNING: " + CHOOSE_PROJECT_AUTOMATICALLY_OPTION + " was set as project, but there was an error retrieving list of user projects. Will use fallback: continue without project.";
                    if (e.getXmlRpcResult() != null)
                        err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                    else {
                        err.println(message + ":");
                        e.printStackTrace(err);
                    }
                }
            }


            final GeniUrn sliceUrn = sliceIdArg.startsWith("urn:") ?
                    GeniUrn.parse(sliceIdArg) :
                    GeniUrn.createGeniUrnFromEncodedParts(
                            userAuth.getNameForUrn() + (projectNameArg == null ? "" : ":" + projectNameArg),
                            "slice",
                            sliceIdArg
                    );
            final boolean sliceExists = existingSlices.contains(sliceUrn);

            if (!silent)
                out.println("slice " + sliceUrn + " " + (sliceExists ? "already exists" : "does not yet exist"));

            //check if project exists
            if (isCreate && !sliceExists && sliceUrn.getEncodedSubAuthName() != null) {
                try {
                    if (projectNames == null)
                        projectNames = sa.getSubAuthorityNames(logger, targetUserUrn);

                    if (!projectNames.contains(sliceUrn.getEncodedSubAuthName()))
                        err.println("WARNING: You do not seem to be a member if the project (\"" + sliceUrn.getEncodedSubAuthName() + "\") in which you wish to create a slice. Projects you are member of: " + projectNames);

                    if (debug)
                        out.println("You are a member of the following projects: " + projectNames);
                } catch (JFedException e) {
                    String message = "Error retrieving list of user projects. (this error will be ignored)";
                    if (e.getXmlRpcResult() != null)
                        err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                    else {
                        err.println(message + ":");
                        e.printStackTrace(err);
                    }
                }
            }

            List<UserSpec> userSpecs = new ArrayList<>();
            if (isCreate) {
                UserSpec localUserSpec = new UserSpec(targetUserUrnString);
                if (sshKeysUserServerKeys) {
                    try {
                        List<String> userKeys = sa.getSshKeysForUser(logger, targetUserUrn);
                        localUserSpec.getSshKey().addAll(userKeys);
                        if (debug)
                            out.println("Added " + userKeys.size() + " keys from server.");
                    } catch (JFedException e) {
                        return callFailed("Error retrieving ssh keys for user " + targetUserUrn, e, err, callLogOutput);
                    }
                }
                if (sshKeysUserCert) {
                    //todo optionally, check for duplicate
                    if (!usesSpeaksFor) {
                        String userKey = KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(user.getPublicKey());
                        localUserSpec.getSshKey().add(userKey);
                        if (debug)
                            out.println("Adding userspec from user certificate.");
                    } else {
                        assert speaksForCredential != null;
                        PublicKey pubKey = null;
                        if (speaksForCredential instanceof AbacCredential) {
                            AbacCredential c = (AbacCredential) speaksForCredential;
                            pubKey = c.getSpokenForPubKey();
                        }
                        if (speaksForCredential instanceof SfaCredential) {
                            SfaCredential c = (SfaCredential) speaksForCredential;
                            if (c.getTargetGid() != null) {
                                Gid gid = new Gid(c.getTargetGid());
                                if (gid.getCertificate() != null)
                                    pubKey = gid.getCertificate().getPublicKey();
                            }
                        }
                        if (pubKey != null) {
                            String userKey = KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(pubKey);
                            localUserSpec.getSshKey().add(userKey);
                            if (debug)
                                out.println("Adding userspec from speaksFor SfaCredential");
                        } else {
                            err.println("WARNING: Could not fetch user public key from user speaksFor certificate. (credential class is " + speaksForCredential.getClass().getName() + ")");
                        }
                    }
                }
                if (sshKeysRspec) {
                    try {
                        ModelRspec modelRspec =
                                new StaxRspecParser(rspec.getXmlString(), BasicRspecFactory.getInstance()).call();
                        List<UserSpec> extraUserSpec = modelRspec.getSshKeys();
                        if (debug)
                            out.println("Adding " + extraUserSpec.size() + " userspecs from RSpec");
                        for (UserSpec us : extraUserSpec) {
                            if (us.getUrn().equals(targetUserUrnString)) {
                                //todo optionally, check for duplicate
                                localUserSpec.getSshKey().addAll(us.getSshKey());
                            } else
                                userSpecs.add(us);
                        }
                    } catch (RspecParseException e) {
                        err.println("WARNING: Error parsing specified RSpec. The users could not be extracted from this RSpec. Note: This tool will continue, and the RSpec will still be used to create slivers.");
                    }
                }
                if (!localUserSpec.getSshKey().isEmpty())
                    userSpecs.add(localUserSpec);
                if (debug) {
                    out.println("The following user specification will be added to all create sliver calls:");
                    for (UserSpec us : userSpecs) {
                        out.println("  User " + us.getUrn() + " with " + us.getSshKey().size() + " keys");
                        for (String key : us.getSshKey())
                            out.println("     " + key);
                    }
                }
            }

            if (!isCreate && !sliceExists) {
                err.println("Error: The slice " + sliceUrn + " does not exist. Cannot continue.");
                return 2;
            }

            if (fake) {
                out.println("Argument processing done. --fake is set, so no further steps are taken.");
                return 0;
            }

            List<AnyCredential> sliceCredentials;
            Date sliceExpirationDate = null;
            if (sliceExists) {
                try {
                    sliceCredentials = sa.getSliceCredentials(logger, sliceUrn);
                    if (!sliceCredentials.isEmpty()) sliceExpirationDate = sliceCredentials.get(0).getExpiresDate();
                    LOG.debug("ExperimenterCli requested slice credentials, and got "+sliceCredentials.size()+" credentials");
                } catch (JFedException e) {
                    return callFailed("Error retrieving slice credential", e, err, callLogOutput);
                }

                if (rspecArg == null) {
                    //no RSpec provided, so get rspecComponentManagers from SLIVER_INFO (or equivalent)
                    try {
                        rspecComponentManagers = sa.getAggregatesForSlice(logger, sliceCredentials, sliceUrn);
                    } catch (JFedException e) {
                        return callFailed("Error retrieving slice credential", e, err, callLogOutput);
                    }
                    if (rspecComponentManagers.size() == 0) {
                        err.println("WARNING: No RSpec was provided using the '--rspec' option, and the SA did not know of any AM's that have resources. " +
                                "Few commands can now do anything. Will try anyway.");
                    }
                }

                if (isRenew || isCreate) {
                    Date currentSliceExpireDate = null;
                    for (AnyCredential credential : sliceCredentials)
                        if (credential.getExpiresDate() != null)
                            currentSliceExpireDate = credential.getExpiresDate();
                    if (currentSliceExpireDate != null && currentSliceExpireDate.before(requestedExpirationDate)) {
                        try {
                            if (debug)
                                out.println("Need to Renew slice, because the slice expires at " + currentSliceExpireDate + " while requested expiration date is " + requestedExpirationDate);
                            if (!silent)
                                out.println("Renewing slice...");
                            List<AnyCredential> newSliceCredentials =
                                    sa.renewSlice(logger, sliceCredentials, requestedExpirationDate);
                            if (newSliceCredentials != null) {
                                sliceCredentials = newSliceCredentials;
                                LOG.debug("ExperimenterCli renewed the slice and got "+sliceCredentials.size()+" credentials");
                                if (!sliceCredentials.isEmpty()) sliceExpirationDate = sliceCredentials.get(0).getExpiresDate();
                            }
                            else
                                out.println("WARNING: did not receive renewed credential");
                        } catch (JFedException e) {
                            return callFailed("Error renewing slice credential", e, err, callLogOutput);
                        }
                    }
                }
            } else {
                assert isCreate;
                if (!createSliceArg) {
                    err.println("Error: The slice " + sliceUrn + " does not exist, and you have not specified the --create-slice option.");
                    return 2;
                }
                UserAndSliceApiWrapper.SliceInfo sliceInfo;
                try {
                    sliceInfo =
                            sa.createSlice(logger, sliceUrn.getEncodedResourceName(), requestedExpirationDate, projectNameArg);
                    if (!sliceInfo.getCredentials().isEmpty())
                        sliceExpirationDate = sliceInfo.getCredentials().get(0).getExpiresDate();
                } catch (JFedException e) {
                    return callFailed("Error creating slice", e, err, callLogOutput);
                }
                sliceCredentials = sliceInfo.getCredentials();
                LOG.debug("ExperimenterCli created a slice, and got "+sliceCredentials.size()+" credentials");
            }
            if (sliceCredentials == null || sliceCredentials.isEmpty()) {
                err.println("Error: Did not receive slice credentials. (but call was successful)");
                return 2;
            }
            LOG.debug("ExperimenterCLI has "+sliceCredentials.size()+" slice credentials");

            if (isCreate && sliceRecoverInfoArg != null) {
                if (debug)
                    out.println("Writing slice recovery info file \"" + sliceRecoverInfoArg + "\"");

                try {
                    FileConfiguration conf = new PropertiesConfiguration();
                    conf.setEncoding("UTF-8");
                    conf.setProperty(KEY_EXPIRATION_TIME,
                            sliceExpirationDate == null ? "unknown"
                                    : ISO_OFFSET_DATE_TIME.format(sliceExpirationDate.toInstant()));
                    conf.setProperty(KEY_REQUEST_RSPEC, rspec.getXmlString());
                    conf.save(new File(sliceRecoverInfoArg));
                } catch (ConfigurationException e) {
                    LOG.error("Error writing slice recovery info to \"" + sliceRecoverInfoArg + "\"", e);
                    err.println("ERROR (non fatal): Exception writing slice recovery info to \"" + sliceRecoverInfoArg + "\": " + e.getMessage());
                }
            }


            if (fakeSliver) {
                out.println("Argument and slice processing done. --fake-sliver is set, so no further steps are taken.");
                return 0;
            }

            List<AggregateManagerWrapper> amWrappers = new ArrayList<>();
            ManifestRspecMerger manifestRspecMerger = new ManifestRspecMerger(ModelRspecType.BASIC);
            for (String rspecComponentManager : rspecComponentManagers) {
                if (!silent)
                    out.println("Contacting  " + rspecComponentManager + "...");

                SfaAuthority currentAuthority =
                        authorityFinder.findByUrn(rspecComponentManager, AuthorityFinder.Purpose.CREATE_SLIVER);
                if (currentAuthority == null) {
                    err.println("Tested authority URN is unknown: \"" + rspecComponentManager + "\"");
                    return 2;
                }
                assert currentAuthority != null;

                final AggregateManagerWrapper am;
                if (amApiVersionArg == null) {
                    am = automaticAggregateManagerWrapperFactory.create(currentAuthority);
                } else {
                    if (amApiVersionArg.equals("2"))
                        am = new AMv2Wrapper(logger, userProvider, conPool, jFedPreferences, currentAuthority);
                    else {
                        assert amApiVersionArg.equals("3");
                        am = new AMv3Wrapper(logger, userProvider, conPool, jFedPreferences, currentAuthority);
                    }
                }
                if (usesSpeaksFor) {
                    am.setSpeaksFor(speakingForUser);
                }
                amWrappers.add(am);

                if (isCreate)
                    try {
                        String requestRspec = rspec.getXmlString();
                        if (rewriteRspecArg) {
                            try {
                                RequestRspecSource rspecSource = new RequestRspecSource(rspec);
                                ModelRspec modelRspec = rspecSource.getModelRspec();
                                RequestRspecSource rspecSource2 = new RequestRspecSource(modelRspec);
                                requestRspec = rspecSource2.getRspecXmlString();
                            } catch (Throwable t) {
                                LOG.error("Error rewriting RSpec", t);
                                err.println("Error parsing and rewriting RSpec, will use original RSpec! Error message: " + t.getMessage());
                            }
                        }

                        String manifest;
                        if (am.hasSeperateAllocateAndProvision()) {
                            AggregateManagerWrapper.AllocateReply allocateReply =
                                    am.allocate(sliceUrn, sliceCredentials, requestRspec, requestedExpirationDate);

                            if (allocateReply != null && allocateReply.getSliverUrns() != null && !allocateReply.getSliverUrns().isEmpty())
                                sa.registerAggregatesForSlice(logger, sliceCredentials, sliceUrn, am.getAmAuthority().getUrn(), allocateReply.getSliverUrns(), requestedExpirationDate );

                            manifest =
                                    am.provisionAndStart(sliceUrn, sliceCredentials, userSpecs, requestedExpirationDate);

                            if (manifest != null) {
                                Set<GeniUrn> sliverUrns = findSliversInManifest(manifest);
                                if (allocateReply != null && allocateReply.getSliverUrns() != null && !allocateReply.getSliverUrns().isEmpty())
                                    sliverUrns.removeAll(allocateReply.getSliverUrns());
                                if (!sliverUrns.isEmpty())
                                    sa.registerAggregatesForSlice(logger, sliceCredentials, sliceUrn, am.getAmAuthority().getUrn(), sliverUrns, requestedExpirationDate);
                            }
                        } else {
                            manifest =
                                    am.createSliver(sliceUrn, sliceCredentials, requestRspec, userSpecs, requestedExpirationDate);

                            if (manifest != null) {
                                Set<GeniUrn> sliverUrns = findSliversInManifest(manifest);
                                sa.registerAggregatesForSlice(logger, sliceCredentials, sliceUrn, am.getAmAuthority().getUrn(), sliverUrns, requestedExpirationDate);
                            }
                        }

                        if (!silent)
                            out.println("Sliver at  " + rspecComponentManager + " is created and initializing...");

                        manifestRspecMerger.setManifest(currentAuthority.getUrn(), new ManifestRspecSource(manifest, ModelRspecType.BASIC));
                    } catch (AggregateManagerWrapper.NotEnoughFreeResourcesException e) {
                        return callFailed("Not enough resources available to create sliver at " + currentAuthority.getUrnString(), e, err, callLogOutput);
                    } catch (JFedException e) {
                        return callFailed("Error creating sliver at " + currentAuthority.getUrnString(), e, err, callLogOutput);
                    }

                if (isDelete)
                    try {
                        am.deleteSliver(sliceUrn, sliceCredentials);

                        if (!silent)
                            out.println("Sliver at  " + rspecComponentManager + " has been deleted.");
                    } catch (JFedException e) {
                        String message =
                                "Error deleting sliver at " + rspecComponentManager + ". (this error will be ignored)";
                        if (e.getXmlRpcResult() != null)
                            err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                        else {
                            err.println(message + ":");
                            e.printStackTrace(err);
                        }
                    }

                if (isStatus)
                    try {
                        LOG.debug("ExperimenterCLI is calling status with "+sliceCredentials.size()+" credentials");
                        StatusDetails statusDetails = am.status(sliceUrn, sliceCredentials);

                        if (!silent)
                            out.println("Sliver at  " + rspecComponentManager + " has status " + statusDetails.getGlobalStatus());
                    } catch (JFedException e) {
                        String message =
                                "Error getting status of sliver at " + rspecComponentManager + ". (this error will be ignored)";
                        if (e.getXmlRpcResult() != null) {
                            err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                            err.println("Request was: " + "\"" + e.getXmlRpcResult().getRequestXmlRpcString() + "\"");
                        } else {
                            err.println(message + ":");
                            e.printStackTrace(err);
                        }
                    }

                if (isRenew)
                    try {
                        am.renewSlivers(sliceUrn, null/*no sliverUrns known*/, sliceCredentials, requestedExpirationDate);

                        if (!silent)
                            out.println("Sliver at  " + rspecComponentManager + " has been renewed successfully");
                    } catch (JFedException e) {
                        String message =
                                "Error renewing sliver at " + rspecComponentManager + ". (this error will be ignored)";
                        if (e.getXmlRpcResult() != null)
                            err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                        else {
                            err.println(message + ":");
                            e.printStackTrace(err);
                        }
                    }
            }

            if (isCreate) {
                if (debug)
                    out.println("Writing combined manifest to file \"" + manifestArg + "\"");
                String combinedManifestXmlString = manifestRspecMerger.getMergedManifest().getRspecXmlString();
                IOUtils.stringToFile(new File(manifestArg), combinedManifestXmlString);

                if (waitUntilReady) {
                    if (!silent)
                        if (rspecComponentManagers.size() == 1)
                            out.println("Will now wait until the sliver is ready...");
                        else
                            out.println("Will now wait until the " + rspecComponentManagers.size() + " slivers are ready...");

                    for (AggregateManagerWrapper am : amWrappers) {
                        SfaAuthority currentAuthority = am.getAmAuthority();
                        String rspecComponentManager = currentAuthority.getUrnString();

                        StatusDetails statusDetails = null;
                        while (statusDetails == null || !statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.READY)) {
                            if (debug)
                                out.println("Waiting 5 seconds before checking status...");
                            try {
                                Thread.sleep(5 * 1000l);
                            } catch (InterruptedException e) {
                                LOG.debug("InterruptedException while sleeping", e);
                            }

                            if (!silent)
                                out.println("Contacting " + rspecComponentManager + " to check status...");

                            try {
                                statusDetails = am.status(sliceUrn, sliceCredentials);
                            } catch (JFedException e) {
                                return callFailed("Error getting sliver status at " + currentAuthority.getUrnString(), e, err, callLogOutput);
                            }

                            if (!silent)
                                out.println("Status of sliver at  " + rspecComponentManager + " is " + statusDetails.getGlobalStatus());

                            if (statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.FAIL) ||
                                    statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.UNALLOCATED)) {
                                return callFailed("ERROR: Sliver status at " + currentAuthority.getUrnString() + " is " +
                                        statusDetails.getGlobalStatus() + " -> " + statusDetails, null, err, callLogOutput);
                            }
                        }
                    }

                    if (!silent)
                        if (rspecComponentManagers.size() == 1)
                            out.println("The sliver is ready.");
                        else
                            out.println("The " + rspecComponentManagers.size() + " slivers are ready.");
                }
            }
        } finally {
            callLogOutput.close();

            if (printCalls) {
                Thread.yield();
                try {
                    Thread.sleep(500l);
                } catch (InterruptedException e) {
                    LOG.error("Unexpected InterruptedException", e);
                }
            }
        }

        if (debug)
            out.println("All done. Exiting.");

        return 0;
    }

    private static int callFailed(String message, JFedException e, PrintStream err, CallLogOutput callLogOutput) {
        Thread.yield();
        Thread.yield();
        callLogOutput.flush();
        try {
            Thread.sleep(100);
        } catch (InterruptedException e1) {
            //ignored
        }
        Thread.yield();
        callLogOutput.flush();

        if (e == null)
            err.println(message);
        else {
            if (e.getXmlRpcResult() != null)
                err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
            else {
                err.println(message + ":");
                e.printStackTrace(err);
            }
        }
        return 2;
    }

    private static class CallLogOutput {
        public FileOutputStream fos;
        public XMLEventWriter xtw;
        public XMLEventFactory eventFactory;
        public final PrintStream err;

        public CallLogOutput(PrintStream err) {
            this.err = err;
        }

        public void onFailure() {
            close();
            fos = null;
            xtw = null;
            eventFactory = null;
        }

        public void flush() {
            try {
                xtw.flush();
                fos.flush();
            } catch (Exception e) {
                LOG.warn("Unexpected error flusing", e);
            }
        }

        public boolean isDisabled() {
            return fos == null || xtw == null || eventFactory == null;
        }

        public void close() {
            if (xtw != null && eventFactory != null) {
                try {
                    xtw.add(eventFactory.createEndElement(new QName("calls"), null));
                    xtw.add(eventFactory.createEndDocument());
                    xtw.flush();
                } catch (XMLStreamException e) {
                    LOG.error("Exception while closing call log output", e);
                    err.println("Exception while closing call log output: " + e.getMessage());
                    e.printStackTrace();
                }
            }
            if (fos != null) {
                try {
                    fos.close();
                } catch (IOException e) {
                    LOG.error("Exception while closing call log output file", e);
                    err.println("Exception while closing call log output file: " + e.getMessage());
                }
            }
        }
    }

    public static Set<GeniUrn> findSliversInManifest(String manifestRspecString) {
        Set<GeniUrn> seenSlivers = new HashSet<>();

        ManifestRspecSource manifestRspecSource = new ManifestRspecSource(manifestRspecString, ModelRspecType.FX);
        ModelRspec rspec = manifestRspecSource.getModelRspec();
        if (rspec == null) return seenSlivers;

        for (RspecNode node : rspec.getNodes())
            if (node.getSliverId() != null)
                seenSlivers.add(node.getSliverId());

        for(RspecLink link : rspec.getLinks())
            if(link.getSliverId() != null)
                seenSlivers.add(link.getSliverId());

        return seenSlivers;
    }
}
