package be.iminds.ilabt.jfed.experimenter_gui.editor.ribbon_tabs;

import be.iminds.ilabt.jfed.experimenter_gui.editor.rspec_validation.RspecCheckReason;
import be.iminds.ilabt.jfed.experimenter_gui.editor.rspec_validation.RspecValidationEngine;
import be.iminds.ilabt.jfed.experimenter_gui.editor.rspec_validation.RspecValidationResult;
import be.iminds.ilabt.jfed.experimenter_gui.editor.views.EditorViewType;
import be.iminds.ilabt.jfed.experimenter_gui.editor.views.RawRspecEditor;
import be.iminds.ilabt.jfed.experimenter_gui.ui.ribbon.RibbonButton;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.fxml.FXML;

/**
 * User: twalcari
 * Date: 7/1/2014
 * Time: 11:10
 */
public class RspecEditorRibbonTab extends EditorRibbonTab {
    private final RspecValidationEngine validationEngine = new RspecValidationEngine();

    @FXML
    private RibbonButton formatCodeButton;
    @FXML
    private RibbonButton verifyCodeButton;
    @FXML
    private RibbonButton searchCodeButton;
    @FXML
    private RibbonButton replaceCodeButton;

    public RspecEditorRibbonTab() {
        FXMLUtil.injectFXML(this, aClass -> {
            if (aClass == GenericEditorRibbonComponentGroupController.class) {
                return new GenericEditorRibbonComponentGroupController(RspecEditorRibbonTab.this);
            }
            throw new RuntimeException("Unexpected class in ControllerFactory: " + aClass.getName());
        });

        setOnSelectionChanged(event -> {
            if (isSelected() && getActiveExperimentEditor().getCurrentEditorViewType() != EditorViewType.RAW) {
                getActiveExperimentEditor().switchToEditorViewByType(EditorViewType.RAW);
            }
        });

    }

    private RawRspecEditor getActiveRawRspecEditor() {
        assert (getActiveExperimentEditor().getCurrentEditorViewType() == EditorViewType.RAW);

        return (RawRspecEditor) getActiveExperimentEditor().getCurrentEditableRspecView();
    }


    @FXML
    private void onFormatCodeAction() {
        assert formatCodeButton != null;
        if (!getActiveRawRspecEditor().formatXmlCode())
            JFDialogs.create()
                    .owner(formatCodeButton)
                    .message("Formatting your Rspec failed. Please check if the XML is valid")
                    .title("The Rspec could not be formatted").showWarning();
    }

    @FXML
    private void onVerifyRspecAction() {
        RequestRspecSource source = getActiveExperimentEditor().getCurrentRequestRspecSource();
        RspecValidationResult verifyRspecResult =
                validationEngine.validateRSpec(RspecCheckReason.CHECK_RSPEC, source, verifyCodeButton.getScene().getWindow());


        assert verifyCodeButton != null;
        if (verifyRspecResult == null) {
            JFDialogs.create()
                    .owner(verifyCodeButton)
                    .title("Rspec Verify Result")
                    .message("No problems found in RSpec.").showInformation();
        } else {
            getActiveExperimentEditor().getEditableRspec().setRequestRspecSource(verifyRspecResult.getReplacementRequestRspecSource());
        }
    }

    @FXML
    private void onSearchAction() {
        getActiveRawRspecEditor().showSearch();
    }

    @FXML
    private void onReplaceAction() {
        getActiveRawRspecEditor().showReplace();
    }

}
