package be.iminds.ilabt.jfed.experimenter_gui.slice.raw;

import be.iminds.ilabt.jfed.experimenter_gui.slice.Experiment;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentChangeListener;
import be.iminds.ilabt.jfed.experimenter_gui.slice.SliceState;
import be.iminds.ilabt.jfed.experimenter_gui.util.BrowserUtil;
import be.iminds.ilabt.jfed.experimenter_gui.util.FileUtils;
import be.iminds.ilabt.jfed.experimenter_gui.util.SSHTerminalUtil;
import be.iminds.ilabt.jfed.experimenter_gui.util.StringConverters;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.rspec.model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.GeantTestbedType;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.XmlUtil;
import be.iminds.jfed.gts_highlevel.controller.GtsModel;
import be.iminds.jfed.gts_highlevel.model.GtsProject;
import be.iminds.jfed.gts_highlevel.model.GtsResource;
import com.github.twalcari.prettify.RTSyntaxHighlighter;
import javafx.application.Platform;
import javafx.beans.InvalidationListener;
import javafx.beans.WeakInvalidationListener;
import javafx.beans.binding.Bindings;
import javafx.beans.binding.DoubleBinding;
import javafx.beans.property.ReadOnlyIntegerWrapper;
import javafx.beans.property.ReadOnlyStringWrapper;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.scene.control.*;
import javafx.scene.control.cell.PropertyValueFactory;
import javafx.scene.control.cell.TextFieldTableCell;
import javafx.scene.input.Clipboard;
import javafx.scene.input.ClipboardContent;
import javafx.scene.layout.HBox;
import javafx.scene.layout.StackPane;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.stage.FileChooser;
import javafx.util.StringConverter;
import org.controlsfx.glyphfont.FontAwesome;
import org.fxmisc.richtext.CodeArea;
import org.fxmisc.richtext.LineNumberFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.time.Instant;
import java.util.*;

/**
 * User: twalcari
 * Date: 2/6/14
 * Time: 8:21 AM
 */
public class RawSliceView extends StackPane implements ExperimentChangeListener {
    private static final Logger LOG = LoggerFactory.getLogger(RawSliceView.class);
    private static final String COMBINED_MANIFEST = "Combined Manifest RSpec";

    private final Experiment experiment;
    private final ManifestRspecSource canvasRspecSource;

    private final SSHTerminalUtil sshTerminalUtil;
    private final BrowserUtil browserUtil;
    private final JFedGuiPreferences jFedPreferences;
    private final GtsModel gtsModel;
    private final StringConverter<Object> manifestComboboxStringConverter = new StringConverter<Object>() {
        @Override
        public String toString(Object object) {
            if (object instanceof SfaAuthority)
                return ((SfaAuthority) object).getHrn();
            else if (object instanceof String)
                return (String) object;
            else
                throw new UnsupportedOperationException("Only String and SfaAuthority are supported");
        }

        @Override
        public Object fromString(String string) {
            throw new UnsupportedOperationException();
        }
    };

    @FXML
    private HBox initializingBox;
    @FXML
    private ScrollPane testbedsScrollPane;
    @FXML
    private VBox testbedsVBox;

    @FXML
    private TableView<Slice> sliceTableView;

    @FXML
    private TableColumn<Slice, String> sliceNameTableColumn;
    @FXML
    private TableColumn<Slice, String> sliceProjectTableColumn;
    @FXML
    private TableColumn<Slice, GeniUrn> sliceURNTableColumn;
    @FXML
    private TableColumn<Slice, Instant> sliceExpirationTimeTableColumn;

    @FXML
    private TableView<Sliver> sliversTableView;

    @FXML
    private TableColumn<Sliver, String> sliverTestbedTableColumn;

    @FXML
    private TableColumn<Sliver, GeniUrn> sliverIdTableColumn;

    @FXML
    private TableColumn<Sliver, Instant> sliverExpirationTimeTableColumn;

    @FXML
    private TableColumn<Sliver, StatusDetails> sliverStatusTableColumn;

    @FXML
    private TableView<NodeWithLoginInfo> nodesTableView;

    @FXML
    private TableColumn<NodeWithLoginInfo, String> nodeNameTableColumn;

    @FXML
    private TableColumn<NodeWithLoginInfo, String> hostnameTableColumn;

    @FXML
    private TableColumn<NodeWithLoginInfo, Number> portTableColumn;

    @FXML
    private TableColumn<NodeWithLoginInfo, String> usernameTableColumn;

    @FXML
    private TableColumn<NodeWithLoginInfo, Boolean> loginTableColumn;

    /**
     * this combobox should only contain String of SfaAuthority-objects!
     */
    @FXML
    private ComboBox<Object> manifestCombobox;
    @FXML
    private Label manifestLabel;

    @FXML
    private CodeArea manifestXmlView;


    private Slice slice = null;

    private File prevSaveFile = null;
    private final RTSyntaxHighlighter highlighter;


    public RawSliceView(Experiment experiment, ManifestRspecSource canvasRspecSource, SSHTerminalUtil sshTerminalUtil,
                        BrowserUtil browserUtil, JFedGuiPreferences jFedPreferences, GtsModel gtsModel) {
        this.experiment = experiment;
        this.canvasRspecSource = canvasRspecSource;
        this.sshTerminalUtil = sshTerminalUtil;
        this.browserUtil = browserUtil;
        this.jFedPreferences = jFedPreferences;
        this.gtsModel = gtsModel;

        FXMLUtil.injectFXML(this);

        testbedsVBox.prefWidthProperty().bind(testbedsScrollPane.widthProperty().subtract(15));

        //initilaize slice tableview
        sliceNameTableColumn.setCellValueFactory(t -> new ReadOnlyStringWrapper(t.getValue().getName()));
        sliceProjectTableColumn.setCellValueFactory(t -> new ReadOnlyStringWrapper(t.getValue().getSubAuthorityName()));
        sliceURNTableColumn.setCellValueFactory(new PropertyValueFactory<>("urn"));
        sliceExpirationTimeTableColumn.setCellValueFactory(new PropertyValueFactory<>("expirationDate"));
        sliceExpirationTimeTableColumn.setCellFactory(TextFieldTableCell.forTableColumn(StringConverters.INSTANT_STRING_CONVERTER));

        DoubleBinding sliceTableWidthLeft = sliceTableView.widthProperty().subtract(2);
        sliceNameTableColumn.prefWidthProperty().bind(sliceTableWidthLeft.multiply(0.15));
        sliceProjectTableColumn.prefWidthProperty().bind(sliceTableWidthLeft.multiply(0.15));
        sliceURNTableColumn.prefWidthProperty().bind(sliceTableWidthLeft.multiply(0.5));
        sliceExpirationTimeTableColumn.prefWidthProperty().bind(sliceTableWidthLeft.multiply(0.2));


        sliceTableView.setRowFactory(param -> {
            TableRow<Slice> row = new TableRow<>();

            ContextMenu sliceUrnContextMenu = new ContextMenu();
            MenuItem copyUrnMenuItem = new MenuItem("Copy URN");
            copyUrnMenuItem.setGraphic(GlyphUtils.createGlyph(FontAwesome.Glyph.COPY));
            copyUrnMenuItem.setOnAction(event -> {
                final Clipboard clipboard = Clipboard.getSystemClipboard();
                final ClipboardContent content = new ClipboardContent();
                content.putString(slice.getUrnString());
                clipboard.setContent(content);
            });
            sliceUrnContextMenu.getItems().add(copyUrnMenuItem);

            row.setContextMenu(sliceUrnContextMenu);
            return row;
        });


        //initialize slivers tableview
        sliverTestbedTableColumn.setCellValueFactory(t -> new ReadOnlyStringWrapper(t.getValue().getAuthority().getName()));

        sliverIdTableColumn.setCellValueFactory(new PropertyValueFactory<>("urn"));


        sliverExpirationTimeTableColumn.setCellValueFactory(new PropertyValueFactory<>("expirationDate"));
        sliverExpirationTimeTableColumn.setCellFactory(TextFieldTableCell.forTableColumn(StringConverters.INSTANT_STRING_CONVERTER));


        sliverStatusTableColumn.setCellValueFactory(new PropertyValueFactory<>("status"));


        sliverStatusTableColumn.setCellFactory(sliverSliverStatusTableColumn -> new StatusTableCell());


        sliversTableView.setRowFactory(param -> {

            TableRow<Sliver> row = new TableRow<>();

            ContextMenu sliverUrnContextMenu = new ContextMenu();
            MenuItem copySliverUrnMenuItem = new MenuItem("Copy URN");
            copySliverUrnMenuItem.setGraphic(GlyphUtils.createGlyph(FontAwesome.Glyph.COPY));
            copySliverUrnMenuItem.setOnAction(event -> {
                final Clipboard clipboard = Clipboard.getSystemClipboard();
                final ClipboardContent content = new ClipboardContent();
                content.putString(row.getItem().getUrnString());
                clipboard.setContent(content);
            });

            sliverUrnContextMenu.getItems().add(copySliverUrnMenuItem);

            row.setContextMenu(sliverUrnContextMenu);
            return row;
        });

        //set table widths
        DoubleBinding sliverTableWidthLeft =
                sliversTableView.widthProperty().subtract(2).subtract(sliverStatusTableColumn.widthProperty());


        sliverTestbedTableColumn.prefWidthProperty().bind(sliverTableWidthLeft.multiply(0.25));
        sliverIdTableColumn.prefWidthProperty().bind(sliverTableWidthLeft.multiply(0.5));
        sliverExpirationTimeTableColumn.prefWidthProperty().bind(sliverTableWidthLeft.multiply(0.25));

        //initialize nodes tableview

        nodeNameTableColumn.setCellValueFactory(t -> new ReadOnlyStringWrapper(t.getValue().getId()));

        hostnameTableColumn.setCellValueFactory(t -> new ReadOnlyStringWrapper(t.getValue().getHostname()));

        portTableColumn.setCellValueFactory(t -> new ReadOnlyIntegerWrapper(t.getValue().getPort()));

        usernameTableColumn.setCellValueFactory(t -> new ReadOnlyStringWrapper(t.getValue().getUsername()));

        loginTableColumn.setCellValueFactory(
                p -> new SimpleBooleanProperty(p.getValue().getHostname() != null)
        );

        loginTableColumn.setCellFactory(t -> new NodeWithLoginInfoButtonTableCell());

        nodesTableView.setRowFactory(param -> {
            TableRow<NodeWithLoginInfo> row = new TableRow<>();

            ContextMenu loginContextManu = new ContextMenu();

            MenuItem copyLoginInfo = new MenuItem("Copy SSH login command");
            copyLoginInfo.setGraphic(GlyphUtils.createGlyph(FontAwesome.Glyph.TERMINAL));
            copyLoginInfo.setOnAction(event -> {
                boolean success = false;

                if (row.getItem() instanceof RspecNodeWithLoginInfo) {
                    RspecNodeWithLoginInfo rspecNodeWithLoginInfo = (RspecNodeWithLoginInfo) row.getItem();
                    final ClipboardContent content = new ClipboardContent();

                    if (rspecNodeWithLoginInfo.getLoginService() != null) {
                        final BasicStringRspec.LoginService loginService = rspecNodeWithLoginInfo.getLoginService();
                        content.putString("ssh " + loginService.getUsername() + "@" + loginService.getHostname() + " -p " + loginService.getPort());
                        Clipboard.getSystemClipboard().setContent(content);
                        success = true;
                    }
                }
                if (!success)
                    new Alert(Alert.AlertType.ERROR, "No SSH info available.").showAndWait();
            });

            loginContextManu.getItems().add(copyLoginInfo);
            loginContextManu.getItems().add(new SeparatorMenuItem());

            MenuItem copyHostnameMenuItem = new MenuItem("Copy hostname");
            copyHostnameMenuItem.setGraphic(GlyphUtils.createGlyph(FontAwesome.Glyph.COPY));
            copyHostnameMenuItem.setOnAction(event -> {
                if (row.getItem().getHostname() != null) {

                    final ClipboardContent content = new ClipboardContent();
                    content.putString(row.getItem().getHostname());
                    Clipboard.getSystemClipboard().setContent(content);

                } else
                    new Alert(Alert.AlertType.ERROR, "No hostname available").showAndWait();
            });

            loginContextManu.getItems().add(copyHostnameMenuItem);

            MenuItem copyUsernameMenuItem = new MenuItem("Copy username");
            copyUsernameMenuItem.setGraphic(GlyphUtils.createGlyph(FontAwesome.Glyph.COPY));
            copyUsernameMenuItem.setOnAction(event -> {


                if (row.getItem().getUsername() != null) {
                    RspecNodeWithLoginInfo rspecNodeWithLoginInfo = (RspecNodeWithLoginInfo) row.getItem();

                    final ClipboardContent content = new ClipboardContent();
                    content.putString(row.getItem().getUsername());
                    Clipboard.getSystemClipboard().setContent(content);
                } else
                    new Alert(Alert.AlertType.ERROR, "No username available").showAndWait();
            });

            loginContextManu.getItems().add(copyUsernameMenuItem);

            MenuItem copyPortMenuItem = new MenuItem("Copy port");
            copyPortMenuItem.setGraphic(GlyphUtils.createGlyph(FontAwesome.Glyph.COPY));
            copyPortMenuItem.setOnAction(event -> {
                if (row.getItem().getPort() != null) {
                    final ClipboardContent content = new ClipboardContent();
                    content.putString(Integer.toString(row.getItem().getPort()));
                    Clipboard.getSystemClipboard().setContent(content);
                } else
                    new Alert(Alert.AlertType.ERROR, "No port available").showAndWait();
            });

            loginContextManu.getItems().add(copyPortMenuItem);


            row.setContextMenu(loginContextManu);
            return row;

        });


        nodesTableView.prefHeightProperty().bind(Bindings.max(100,
                Bindings.size(nodesTableView.getItems()).multiply(28).add(25 /*header */ + 2 /*safety margin */)));

        //set tablecolumn widths
        nodeNameTableColumn.prefWidthProperty().bind(
                nodesTableView.widthProperty().subtract(loginTableColumn.getPrefWidth() + 2).multiply(0.3));
        hostnameTableColumn.prefWidthProperty().bind(
                nodesTableView.widthProperty().subtract(loginTableColumn.getPrefWidth() + 2).multiply(0.3));
        portTableColumn.prefWidthProperty().bind(
                nodesTableView.widthProperty().subtract(loginTableColumn.getPrefWidth() + 2).multiply(0.2));
        usernameTableColumn.prefWidthProperty().bind(
                nodesTableView.widthProperty().subtract(loginTableColumn.getPrefWidth() + 2).multiply(0.2));

        visibleProperty().addListener(observable -> {
            if (isVisible() && slice != null) {
                sliversTableView.itemsProperty().unbind();
                //sliversTableView.getItems().clear();
                sliversTableView.setItems(slice.getSlivers());

            }
        });

        manifestCombobox.setConverter(manifestComboboxStringConverter);
        manifestCombobox.getItems().add(COMBINED_MANIFEST);
        manifestCombobox.getSelectionModel().selectFirst();
        manifestCombobox.getSelectionModel().selectedItemProperty().addListener(observable -> updateManifestView());

        manifestXmlView.setParagraphGraphicFactory(LineNumberFactory.get(manifestXmlView));
        manifestXmlView.prefWidthProperty().bind(testbedsVBox.widthProperty().subtract(42));
        highlighter = new RTSyntaxHighlighter(manifestXmlView, "xml");

        experiment.addExperimentChangeListener(this);
        if (experiment.getSlice() != null) {
            this.initializeSlice(experiment.getSlice());
        }

        gtsModel.getResources().addListener((InvalidationListener) observable -> updateGtsReservations());
    }

    private final InvalidationListener manifestInvalidationListener = observable -> {
        if (Platform.isFxApplicationThread()) {
            updateView();
        } else {
            Platform.runLater(this::updateView);
        }
    };

    private void disableInitializingView() {
        initializingBox.setVisible(false);
        initializingBox.setManaged(false);
        testbedsScrollPane.setVisible(true);
    }

    private void updateNodesTable() {
        ObservableList<NodeWithLoginInfo> nodeWithLoginInfos = FXCollections.observableArrayList();
        nodeWithLoginInfos.addAll(gtsNodesInfo);

        if (slice.getManifestRspec() != null && slice.getManifestRspec().getModelRspec(ModelRspecType.FX) != null) {

            for (FXRspecNode node : ((FXModelRspec) slice.getManifestRspec().getModelRspec(ModelRspecType.FX)).getNodes()) {
                List<BasicStringRspec.LoginService> loginServices = slice.getManifestRspec().getNodeLoginInfo(node);
                if (loginServices == null || loginServices.isEmpty()) {
                    nodeWithLoginInfos.add(new RspecNodeWithLoginInfo(node, null));
                } else {
                    for (BasicStringRspec.LoginService loginService : loginServices) {
                        nodeWithLoginInfos.add(new RspecNodeWithLoginInfo(node, loginService));
                    }
                }
            }

            FXCollections.sort(nodeWithLoginInfos, Comparator.comparing(NodeWithLoginInfo::getId)
                    .thenComparing(Comparator.nullsLast(Comparator.comparing(NodeWithLoginInfo::getHostname)))
                    .thenComparing(Comparator.nullsLast(Comparator.comparing(NodeWithLoginInfo::getPort)))
                    .thenComparing(Comparator.nullsLast(Comparator.comparing(NodeWithLoginInfo::getUsername))));

        } else {
            LOG.warn("Could not retrieve FXModelRspec from slice.getManifestRspec()!");
        }

        nodesTableView.getItems().setAll(nodeWithLoginInfos);
    }

    private void updateView() {
        //if we don't have a manifest, we're still initializing
        if (slice.getManifestRspec() == null)
            return;

        disableInitializingView();

        updateNodesTable();
        //update the available manifest rspecs:

        Object oldSelection = manifestCombobox.getSelectionModel().getSelectedItem();

        List<SfaAuthority> manifestAMsList = new ArrayList<>(slice.getManifestComponentManagers());
        manifestAMsList.sort(Comparator.comparing(SfaAuthority::getName));

        ObservableList<Object> allManifests = FXCollections.observableArrayList();
        allManifests.add(COMBINED_MANIFEST);
        allManifests.addAll(manifestAMsList);

        manifestCombobox.setItems(allManifests);
        manifestCombobox.getSelectionModel().select(oldSelection);

        updateManifestView();
    }

    @FXML
    private void updateManifestView() {
        //if we don't have a manifest, we're still initializing
        if (slice == null || slice.getManifestRspec() == null)
            return;

        assert Platform.isFxApplicationThread();

        ManifestRspecSource manifestRspecSource = slice.getManifestRspec();

        if (!manifestCombobox.getSelectionModel().isEmpty() &&
                manifestCombobox.getSelectionModel().getSelectedItem() != COMBINED_MANIFEST) {
            assert manifestCombobox.getSelectionModel().getSelectedItem() instanceof SfaAuthority;
            manifestRspecSource = slice.getSpecificManifestRspec(
                    ((SfaAuthority) manifestCombobox.getSelectionModel().getSelectedItem()));
        }

        if (manifestRspecSource != null) {

            manifestXmlView.replaceText(XmlUtil.formatXmlFromString(manifestRspecSource.getRspecXmlString()));
        } else {
            manifestXmlView.replaceText("No manifest available");
        }
    }

    @FXML
    private void saveManifest() {
        //if we don't have a manifest, we're still initializing
        if (slice == null || slice.getManifestRspec() == null)
            return;

        assert Platform.isFxApplicationThread();
        ManifestRspecSource manifestRspecSource = slice.getManifestRspec();

        if (manifestCombobox.getSelectionModel().getSelectedItem() != COMBINED_MANIFEST) {
            assert manifestCombobox.getSelectionModel().getSelectedItem() instanceof SfaAuthority;
            manifestRspecSource = slice.getSpecificManifestRspec(
                    ((SfaAuthority) manifestCombobox.getSelectionModel().getSelectedItem()));
        }

        if (manifestRspecSource == null) {
            LOG.warn("No manifest available to save");
            return;
        }

        String selectedTitleString =
                manifestComboboxStringConverter.toString(manifestCombobox.getSelectionModel().getSelectedItem());
        String manifestXmlString = XmlUtil.formatXmlFromString(manifestRspecSource.getRspecXmlString());

        FileChooser fc = new FileChooser();
        fc.setTitle("Save Manifest RSpec");
        fc.getExtensionFilters().add(FileUtils.MANIFEST_RSPEC_EXTENSION_FILTER);
        fc.setInitialFileName(selectedTitleString + FileUtils.MANIFEST_RSPEC_EXTENSION);


        //set the default save dir
        if (System.getProperty("user.home") != null)
            fc.setInitialDirectory(new File(System.getProperty("user.home")));

        if (jFedPreferences.containsPreference(GuiPreferenceKey.PREF_RSPECFOLDER_MANIFEST)) {
            File dir = new File(jFedPreferences.getString(GuiPreferenceKey.PREF_RSPECFOLDER_MANIFEST));
            if (dir.exists()) {
                fc.setInitialDirectory(dir);
            }
        } else {
            if (jFedPreferences.containsPreference(GuiPreferenceKey.PREF_RSPECFOLDER_MANIFEST)) {
                File dir = new File(jFedPreferences.getString(GuiPreferenceKey.PREF_RSPECFOLDER_REQUEST));
                if (dir.exists()) {
                    fc.setInitialDirectory(dir);
                }
            }
        }


        if (prevSaveFile != null) {
            fc.setInitialDirectory(prevSaveFile.getParentFile());
//            fc.setInitialFileName(prevSaveFile.getName());
        }

        File saveFile = fc.showSaveDialog(this.getScene().getWindow());

        if (saveFile == null)
            return;

        //save folder in which the file was saved for future file selector dialogs
        jFedPreferences.setString(GuiPreferenceKey.PREF_RSPECFOLDER_MANIFEST,
                saveFile.getParentFile().getAbsolutePath());

        if (!saveFile.getAbsolutePath().endsWith(FileUtils.MANIFEST_RSPEC_EXTENSION))
            saveFile = new File(saveFile.getAbsolutePath() + FileUtils.MANIFEST_RSPEC_EXTENSION);

        final String fileContents = manifestXmlString;
        final File finalSaveFile = saveFile;
        final javafx.concurrent.Task<Void> saveTask = new javafx.concurrent.Task<Void>() {

            @Override
            protected Void call() throws Exception {
                updateMessage("Saving manifest in \"" + finalSaveFile.getAbsolutePath() + "\"");

                try (FileWriter writer = new FileWriter(finalSaveFile)) {
                    writer.write(fileContents);
                    writer.flush();
                    writer.close();
                } catch (final IOException ex) {
                    LOG.error("Error while saving experiment to file '" + finalSaveFile.getAbsolutePath() + "'", ex);
                    Platform.runLater(new Runnable() {
                        @Override
                        public void run() {
                            JFDialogs.create()
                                    .owner(this)
                                    .message("An error occurred while saving the rspec.")
                                    .masthead("Error")
                                    .title("Error")
                                    .showException(ex);

                        }
                    });
                }
                updateMessage("Successfully saved experiment in \"" + finalSaveFile.getAbsolutePath() + "\"");

                return null;
            }
        };
        //saveTask.messageProperty().addListener(observable -> setStatus(saveTask.getMessage()));

        new Thread(saveTask).start();
    }

    public Slice getSlice() {
        return slice;
    }

    private void initializeSlice(Slice slice) {
        // a slice can only be set one time, after  that, the field becomes immutable
        assert (this.slice == null);
        this.slice = slice;

        LOG.debug("Initializing RawSliceView with slice {}", slice.getUrn());

        Platform.runLater(() -> {
            sliceTableView.getItems().add(slice);

            sliversTableView.setItems(slice.getSlivers());

            sliversTableView.prefHeightProperty().bind(Bindings.max(100,
                    Bindings.size(sliversTableView.getItems()).multiply(25).add(25 /*header*/ + 2 /* safety margin */)));

            slice.manifestRspecProperty().addListener(new WeakInvalidationListener(manifestInvalidationListener));
            updateView();
        });
    }

    @Override
    public void onSliceStateChange(SliceState newState) {
        if (this.slice == null && newState != SliceState.PENDING) {
            assert (experiment.getSlice() != null);
            initializeSlice(experiment.getSlice());
        }
    }

    @Override
    public void onGtsReservationAdded(String geantTestbedTypeName, Integer reservationId) {
        //TODO


    }

    private List<GtsNodeWithLoginInfo> gtsNodesInfo = Collections.emptyList();

    private void updateGtsReservations() {
        FXModelRspec modelRspec = (FXModelRspec) canvasRspecSource.getModelRspec(ModelRspecType.FX);
        assert modelRspec != null;

        List<GtsNodeWithLoginInfo> newGtsNodeInfoList = new ArrayList<>();
        for (Map.Entry<String, Integer> gtsReservationEntry : experiment.getGtsReservationIds().entrySet()) {

            GeantTestbedType gtt = modelRspec.getGeantTestbedTypeByName(gtsReservationEntry.getKey());

            for (GtsResource resource : gtsModel.getResources().values()) {
                boolean isInReservation =
                        resource.getReservations().stream().anyMatch(reservation -> Objects.equals(reservation.getId(), gtsReservationEntry.getValue()));
                if (isInReservation) {
                    URL vncUrl = null;
                    if (resource.getCapabilities().containsKey("console")) {

                        try {
                            vncUrl = new URL(resource.getCapabilities().get("console"));
                        } catch (MalformedURLException e) {
                            LOG.warn("Got invalid URL {} while processing GTS resource {}",
                                    resource.getCapabilities().get("console"), resource.getId(), e);
                        }

                    }
                    newGtsNodeInfoList.add(new GtsNodeWithLoginInfo(gtt, resource.getId(), vncUrl));
                }
            }
        }

        gtsNodesInfo = newGtsNodeInfoList;
        updateNodesTable();
    }


    @Override
    public void onGtsProjectRegistered(GtsProject project) {
        disableInitializingView();
        //TODO
    }

    private static class StatusTableCell extends TableCell<Sliver, StatusDetails> {
        private static String statusToStyleClass(StatusDetails status) {
            return statusToStyleClass(status.getGlobalStatus());
        }

        private static String statusToStyleClass(StatusDetails.SliverStatus status) {
            switch (status) {
                case READY:
                    return "testbed-status-ready";

                case CHANGING:
                    return "testbed-status-changing";

                case FAIL:
                    return "testbed-status-fail";

                case UNALLOCATED:
                    return "testbed-status-unallocated";

                case UNINITIALISED:
                    return "testbed-status-uninitialised";

                case UNKNOWN:
                    return "testbed-status-unknown";
                default:
                    throw new IllegalArgumentException();
            }
        }

        @Override
        protected void updateItem(StatusDetails item, boolean empty) {
            super.updateItem(item, empty);

            if (LOG.isTraceEnabled())
                LOG.trace("Update sliver status with {} and {}", item, empty);

            for (StatusDetails.SliverStatus s : StatusDetails.SliverStatus.values()) {
                getStyleClass().remove(statusToStyleClass(s));
            }

            if (empty) {
                setText("");
            } else {
                setText(item.getGlobalStatus().name());
                getStyleClass().add(statusToStyleClass(item));
            }
        }
    }

    private interface NodeWithLoginInfo {
        String getId();

        String getHostname();

        Integer getPort();

        String getUsername();
    }


    private static class RspecNodeWithLoginInfo implements NodeWithLoginInfo {
        private final FXRspecNode rspecNode;
        private final BasicStringRspec.LoginService loginService;

        private RspecNodeWithLoginInfo(FXRspecNode rspecNode, BasicStringRspec.LoginService loginService) {
            this.rspecNode = rspecNode;
            this.loginService = loginService;
        }

        public String getId() {
            return rspecNode.getClientId();
        }

        @Override
        public String getHostname() {
            return loginService != null ? loginService.getHostname() : null;
        }

        @Override
        public Integer getPort() {
            return loginService != null ? loginService.getPort() : null;
        }

        @Override
        public String getUsername() {
            return loginService != null ? loginService.getUsername() : null;
        }

        public FXRspecNode getRspecNode() {
            return rspecNode;
        }

        public BasicStringRspec.LoginService getLoginService() {
            return loginService;
        }
    }

    private static class GtsNodeWithLoginInfo implements NodeWithLoginInfo {
        private final GeantTestbedType gtt;
        private final String hostId;
        private final URL vncUrl;

        private GtsNodeWithLoginInfo(GeantTestbedType gtt, String hostId, URL vncUrl) {
            this.gtt = gtt;
            this.hostId = hostId;
            this.vncUrl = vncUrl;
        }

        public URL getVncUrl() {
            return vncUrl;
        }

        @Override
        public String getId() {
            return "[GTS " + gtt.getName() + "] " + hostId;
        }

        @Override
        public String getHostname() {
            return null;
        }

        @Override
        public Integer getPort() {
            return null;
        }

        @Override
        public String getUsername() {
            return null;
        }
    }

    private class NodeWithLoginInfoButtonTableCell extends TableCell<NodeWithLoginInfo, Boolean> {

        private final Button loginButton;

        public NodeWithLoginInfoButtonTableCell() {
            loginButton = new Button("Login");

            loginButton.setGraphic(GlyphUtils.createSmallGlyph(FontAwesome.Glyph.TERMINAL, Color.BLACK));
            setGraphic(loginButton);
            setContentDisplay(ContentDisplay.GRAPHIC_ONLY);
        }

        @Override
        protected void updateItem(Boolean item, boolean empty) {
            if (empty) {
                setText(null);
                setGraphic(null);
            } else {
                if (getTableRow().getItem() instanceof RspecNodeWithLoginInfo) {
                    RspecNodeWithLoginInfo rspecNodeWithLoginInfo = (RspecNodeWithLoginInfo) getTableRow().getItem();

                    loginButton.setDisable(!item);
                    loginButton.setOnAction(actionEvent -> sshTerminalUtil.launchSSHTerminal(rspecNodeWithLoginInfo.getLoginService()));
                } else if (getTableRow().getItem() instanceof GtsNodeWithLoginInfo) {
                    GtsNodeWithLoginInfo gtsNodeWithLoginInfo = (GtsNodeWithLoginInfo) getTableRow().getItem();

                    loginButton.setDisable(gtsNodeWithLoginInfo.getVncUrl() == null);
                    loginButton.setOnAction(actionEvent -> browserUtil.openUrlInBrowser(gtsNodeWithLoginInfo.getVncUrl()));
                }
            }
        }
    }
}
