package be.iminds.ilabt.jfed.experimenter_gui.slice.tasks;

import be.iminds.ilabt.jfed.connectivity_tester.ConnectivityTest;
import be.iminds.ilabt.jfed.experimenter_gui.bugreporting.BugReportDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.SliceState;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.jobs.InstantStartSliversJob;
import be.iminds.ilabt.jfed.highlevel.jobs.InstantStartSliversJobFactory;
import be.iminds.ilabt.jfed.highlevel.jobs.JobExecutor;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.tasks.*;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.GeantTestbedType;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.jfed.gts_highlevel.jobs.GtsJobFactory;
import be.iminds.jfed.gts_highlevel.jobs.GtsStartTestbedsJob;
import be.iminds.jfed.gts_highlevel.model.GtsReservation;
import be.iminds.jfed.gts_highlevel.tasks.CreateGtsProjectTask;
import be.iminds.jfed.gts_highlevel.tasks.GtsLoginTask;
import be.iminds.jfed.gts_highlevel.tasks.StartGtsResourcesTask;
import be.iminds.jfed.gts_highlevel.tasks.UpdateGtsReservationsUntilReadyTask;
import javafx.application.HostServices;
import javafx.application.Platform;
import javafx.collections.ListChangeListener;
import javafx.collections.MapChangeListener;
import javafx.scene.control.*;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.scene.text.Text;
import javafx.scene.text.TextFlow;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Provider;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 2/3/2015
 * Time: 07:40
 */
public class InitializeSliversWithoutStitchingTaskGenerator extends InitializeSliversTaskGenerator {
    private static final Logger LOG = LoggerFactory.getLogger(InitializeSliversWithoutStitchingTaskGenerator.class);

    private final ExperimentTasksFactory experimentTasksFactory;
    private final JFedPreferences jFedPreferences;
    private final Provider<HostServices> hostServicesProvider;
    private final BugReportDialogFactory bugReportDialogFactory;

    private final InstantStartSliversJobFactory instantStartSliversJobFactory;
    private final GtsJobFactory gtsJobFactory;
    private final JobExecutor jobExecutor;
    private final ListChangeListener<TaskExecution> taskExecutionListChangeListener =
            (ListChangeListener<TaskExecution>) c -> {
                while (c.next()) {
                    if (c.wasAdded()) {
                        c.getAddedSubList().forEach(this::onNewTaskExecution);
                    }
                }
            };


    InitializeSliversWithoutStitchingTaskGenerator(ExperimentController experimentController, HighLevelTaskFactory hltf,
                                                   ExperimentTasksFactory experimentTasksFactory, JFedPreferences jFedPreferences,
                                                   Provider<HostServices> hostServicesProvider, BugReportDialogFactory bugReportDialogFactory,
                                                   InstantStartSliversJobFactory instantStartSliversJobFactory,
                                                   GtsJobFactory gtsJobFactory,
                                                   JobExecutor jobExecutor) {
        super(experimentController, hltf);
        this.experimentTasksFactory = experimentTasksFactory;
        this.jFedPreferences = jFedPreferences;
        this.hostServicesProvider = hostServicesProvider;
        this.bugReportDialogFactory = bugReportDialogFactory;
        this.instantStartSliversJobFactory = instantStartSliversJobFactory;
        this.gtsJobFactory = gtsJobFactory;
        this.jobExecutor = jobExecutor;
    }

    @Override
    public TaskStatusIndicator.Status generateExperimentTasks() {
        assert (experiment.getSlice() != null);
        assert (experiment.getRequestedEndTime() != null);

        //check if there are GTS-resources to provision

        assert experiment.getSlice().getRequestRspec() != null;
        FXModelRspec modelRspec =
                (FXModelRspec) experiment.getSlice().getRequestRspec().getModelRspec(ModelRspecType.FX);

        if (modelRspec != null && !modelRspec.getGeantTestbedTypes().isEmpty()) {

            GtsStartTestbedsJob gtsStartTestbedsJob =
                    gtsJobFactory.createGtsStartTestbedsJob(experiment.getName(), experiment.getSlice().getRequestRspec());

            gtsStartTestbedsJob.getTaskExecutions().addListener(taskExecutionListChangeListener);

            //listen to results of gts start job
            gtsStartTestbedsJob.projectProperty().addListener(observable -> experiment.setProject(gtsStartTestbedsJob.getProject()));
            gtsStartTestbedsJob.getReservations().addListener((MapChangeListener<GeantTestbedType, GtsReservation>) change -> {
                if (change.wasAdded()) {
                    experiment.getGtsReservationIds().put(change.getKey().getName(), change.getValueAdded().getId());
                } else {
                    LOG.warn("Unexpected change of reservations map: {}", change);
                }
            });

            Future<Void> jobFuture = jobExecutor.submitJob(gtsStartTestbedsJob);

            try {
                jobFuture.get();
            } catch (InterruptedException | ExecutionException e) {
                LOG.error("Exception while starting GTS resources: {}", e.getMessage(), e);
                return TaskStatusIndicator.Status.FAILED;
            }
        }


        InstantStartSliversJob instantStartSliversJob =
                instantStartSliversJobFactory.create(experiment.getSlice(), experiment.getRequestedEndTime(),
                        experiment.isIncludeProjectMemberSshKeys());


        instantStartSliversJob.getTaskExecutions().addListener(taskExecutionListChangeListener);

        Future<Void> jobFuture = jobExecutor.submitJob(instantStartSliversJob);

        try {
            jobFuture.get();    //wait for the job to finish


            if (experiment.getSliceState() == SliceState.PROVISIONING) {
                for (Map.Entry<SfaAuthority, InternalState> authorityInternalStateEntry : experiment.getSlice().getAuthorityStates().entrySet()) {

                    if (authorityInternalStateEntry.getValue() == InternalState.FAILED) {
                        LOG.error("Setting slice state to FAILED because internal state of authority {} is FAILED",
                                authorityInternalStateEntry.getKey());
                        experiment.setSliceState(SliceState.FAILED);
                        return null;
                    }

                    assert authorityInternalStateEntry.getValue() == InternalState.READY;

                }
                experiment.setSliceState(SliceState.READY);
            } else
                LOG.warn("Cannot transfer to SliceState READY as current state is {} instead of PROVISIONING", experiment.getSliceState());

        } catch (InterruptedException ex) {
            LOG.warn("Waiting for InstantStartSliversJob was interrupted", ex);
        } catch (ExecutionException ex) {
            LOG.error("Error while executing InstantStartSliversJob", ex);
        }
        return null;

    }

    @SuppressWarnings("unchecked")
    private void onNewTaskExecution(TaskExecution taskExecution) {

        if (taskExecution.getTask() instanceof CreateSliverReservationTask) {
            experimentController.submitExperimentTask(
                    new RunningCallExperimentTask(
                            String.format("Reserving resources at %s",
                                    ((TaskExecution<CreateSliverReservationTask>) taskExecution)
                                            .getTask().getAuthority().getHrn()),
                            experiment,
                            taskExecution
                    ));


        } else if (taskExecution.getTask() instanceof CreateSliverTask) {
            experimentController.submitExperimentTask(
                    new CreateSliverExperimentTask(taskExecution));

        } else if (taskExecution.getTask() instanceof AllocateSliverTask) {
            experimentController.submitExperimentTask(
                    new AllocateSliverExperimentTask(taskExecution));


        } else if (taskExecution.getTask() instanceof ProvisionSliverTask) {
            experimentController.submitExperimentTask(
                    new RunningCallExperimentTask(
                            String.format("Provisioning resources at %s",
                                    ((TaskExecution<ProvisionSliverTask>) taskExecution)
                                            .getTask().getAuthority().getHrn()),
                            experiment,
                            taskExecution
                    ));

        } else if (taskExecution.getTask() instanceof GetStatusUntilReadyOrFailTask) {
            final GetStatusUntilReadyOrFailTask statusTask =
                    (GetStatusUntilReadyOrFailTask) taskExecution.getTask();
            StatusUntilReadyTask statusUntilReadyTask = new StatusUntilReadyTask(experiment, statusTask);


            statusUntilReadyTask.addTaskFinishedCallback(status -> {
                if (status == TaskStatusIndicator.Status.SUCCESS) {
                    tryStartConnectivityTest(statusTask.getAuthority());

                } else {
                    Platform.runLater(() -> JFDialogs.create()
                            .owner(experimentController.getParentWindow())
                            .message(String.format("One or more resources on %s have permanently failed. " +
                                            "Consider terminating this experiment and trying again. " +
                                            "If the problem continues, please contact support to investigate the issue.",
                                    statusTask.getAuthority().getHrn()))
                            .masthead("Permanent failure on " + statusTask.getAuthority().getHrn())
                            .title("Permanent failure on " + statusTask.getAuthority().getHrn())
                            .showError());
                }
            });

            experimentController.submitExperimentTask(statusUntilReadyTask);

        } else if (taskExecution.getTask() instanceof ListSliceResourcesTask) {
            experimentController.submitExperimentTask(
                    new RunningCallExperimentTask(
                            String.format("Fetching resources description at %s",
                                    ((TaskExecution<ListSliceResourcesTask>) taskExecution)
                                            .getTask().getAuthority().getHrn()),
                            experiment,
                            taskExecution
                    ));

        } else if (taskExecution.getTask() instanceof DeleteSliversAtAuthorityTask) {
            experimentController.submitExperimentTask(
                    new RunningCallExperimentTask(
                            String.format("Performing rollback: deleting resources at %s",
                                    ((TaskExecution<DeleteSliversAtAuthorityTask>) taskExecution)
                                            .getTask().getAuthority().getHrn()),
                            experiment,
                            taskExecution
                    ));
        } else if (taskExecution.getTask() instanceof GtsLoginTask) {
            experimentController.submitExperimentTask(new RunningCallExperimentTask<>(
                    "Logging in on GTS", experiment, taskExecution));

        } else if (taskExecution.getTask() instanceof CreateGtsProjectTask) {
            experimentController.submitExperimentTask(
                    new RunningCallExperimentTask<CreateGtsProjectTask>("Creating GTS Project",
                            experiment,
                            taskExecution
                    ) {
                        @Override
                        public void onTaskFinished(TaskExecution taskExecution) {
                            super.onTaskFinished(taskExecution);

                            if (taskExecution.getState() == TaskExecution.TaskState.SUCCESS) {
                                updateMessage("Created project '" + task.getProject().getId() + "'");
                            }
                        }
                    });
        } else if (taskExecution.getTask() instanceof StartGtsResourcesTask) {
            experimentController.submitExperimentTask(
                    new RunningCallExperimentTask<StartGtsResourcesTask>("Starting GTS Resource '" +
                            ((StartGtsResourcesTask) taskExecution.getTask()).getScriptName() + "'",
                            experiment,
                            taskExecution
                    ));
        } else if (taskExecution.getTask() instanceof UpdateGtsReservationsUntilReadyTask) {
            experimentController.submitExperimentTask(
                    new GtsStatusUntilReadyTask(experiment,
                            (UpdateGtsReservationsUntilReadyTask) taskExecution.getTask()));
        } else {
            LOG.warn("Unexpected task-execution: {}", taskExecution);
        }
    }

    private void tryStartConnectivityTest(SfaAuthority authority) {
        Collection<FXRspecNode> nodesToTest =
                AuthorityConnectivityTestTask.getNodesToTest(experiment.getSlice(), authority);
        if (nodesToTest != null) {
            final AuthorityConnectivityTestTask authorityConnectivityTestTask =
                    experimentTasksFactory.createAuthorityConnectivityTestTask(experiment, authority, nodesToTest);
            authorityConnectivityTestTask.addTaskFinishedCallback(status -> {

                //check if we should show a warning about connectivity issues
                final List<String> failedNodeNames =
                        authorityConnectivityTestTask.getNodesConnectivity().entrySet().stream()
                                .filter(entry -> entry.getValue() == ConnectivityTest.Status.FAILED)
                                .map(entry -> entry.getKey().getClientId())
                                .collect(Collectors.toList());


                Platform.runLater(() -> {
                    if (!failedNodeNames.isEmpty()) {
                        //show a (mutable) notification to mac-users that the terminal opens in the background
                        if (jFedPreferences.getBoolean(GuiPreferenceKey.PREF_SHOW_CONNECTIVITY_WARNING_DIALOG, true)) {
                            LOG.debug("Showing connectivity warning dialog");

                            VBox pane = new VBox();
                            pane.setSpacing(20);

                            final Text introText =
                                    new Text(String.format("jFed detected that node(s) %s cannot be reached.\n" +
                                            "You can take the following steps to resolve the problem:\n\n", failedNodeNames.stream().collect(Collectors.joining(", "))));
                            introText.setStyle("-fx-font-weight: bold");

                            final Hyperlink text1 =
                                    new Hyperlink("1. Try enabling the proxy for SSH connections\n");
                            text1.setOnAction(actionEvent ->
                                    hostServicesProvider.get()
                                            .showDocument("http://doc.ilabt.iminds.be/jfed-documentation/advancedfeatures.html#proxy-settings"));

                            final Hyperlink text2 = new Hyperlink("2. Check the support forum");
                            text2.setOnAction(actionEvent ->
                                    hostServicesProvider.get().showDocument("http://doc.fed4fire.eu/support.html"));

                            final Hyperlink text3 =
                                    new Hyperlink("3. Request assistance by submitting a bug report");
                            text3.setOnAction(actionEvent -> bugReportDialogFactory.showDialog());

                            TextFlow errorMessageTextFlow = new TextFlow(
                                    introText,
                                    text1, new Text("\n"),
                                    text2, new Text("\n"),
                                    text3
                            );

                            pane.getChildren().add(errorMessageTextFlow);
                            final CheckBox checkBox = new CheckBox("Do not show this dialog again");
                            pane.getChildren().add(checkBox);

                            DialogPane dialogPane = new DialogPane();
                            dialogPane.setHeaderText("Experiment connectivity issues");
                            dialogPane.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.PLAY, Color.GREEN));
                            dialogPane.setContent(pane);
                            dialogPane.getButtonTypes().setAll(ButtonType.OK);


                            Dialog<ButtonType> dialog = new Dialog<>();
                            dialog.setTitle("Experiment connectivity issues");
                            dialog.setDialogPane(dialogPane);

                            dialog.showAndWait();

                            jFedPreferences.setBoolean(GuiPreferenceKey.PREF_SHOW_CONNECTIVITY_WARNING_DIALOG,
                                    !checkBox.isSelected());
                        } else {
                            LOG.debug("Skipping connectivity issues dialog because user muted it before");
                        }
                    }
                });

            });

            experimentController.submitExperimentTask(authorityConnectivityTestTask);
        }
    }


    private class CreateSliverExperimentTask extends RunningCallExperimentTask<CreateSliverTask> {
        public CreateSliverExperimentTask(TaskExecution<CreateSliverTask> taskExecution) {
            super(String.format("Allocating resources at %s", taskExecution.getTask().getAuthority().getHrn()),
                    InitializeSliversWithoutStitchingTaskGenerator.this.experiment, taskExecution);
        }

        @Override
        public void onTaskFinished(TaskExecution taskExecution) {
            if (taskExecution.getState() == TaskExecution.TaskState.SUCCESS) {
                updateMessage("Provisioning was also completed");
            } else {
                if (getTask().hasFailedDueToNotEnoughFreeResources()) {
                    Platform.runLater(() -> {
                        updateMessage("Could not map to resources (resource shortage or unsupported topology))");
                        JFDialogs.create()
                                .owner(experimentController.getParentWindow())
                                .message(String.format("Initializing resources at %s failed, because the request could not be mapped on the resources available on the testbed. " +
                                                "This is probably caused by a resource shortage, or because the requested topology is not supported",
                                        getTask().getAuthority().getName()))
                                .masthead("Initializing resources failed")
                                .showError();
                    });
                }
            }
        }
    }

    private class AllocateSliverExperimentTask extends RunningCallExperimentTask<AllocateSliverTask> {
        public AllocateSliverExperimentTask(TaskExecution<AllocateSliverTask> taskExecution) {
            super(String.format("Allocating resources at %s", taskExecution.getTask().getAuthority().getHrn()),
                    InitializeSliversWithoutStitchingTaskGenerator.this.experiment, taskExecution);
        }

        @Override
        public void onTaskFinished(TaskExecution taskExecution) {
            if (taskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
                if (getTask().hasFailedDueToNotEnoughFreeResources()) {
                    Platform.runLater(() -> {
                        updateMessage("Could not map to resources (resource shortage or unsupported topology))");
                        JFDialogs.create()
                                .owner(experimentController.getParentWindow())
                                .message(String.format("Initializing resources at %s failed, because the request could not be mapped on the resources available on the testbed. " +
                                                "This is probably caused by a resource shortage, or because the requested topology is not supported",
                                        getTask().getAuthority().getName()))
                                .masthead("Initializing resources failed")
                                .showError();
                    });
                }
            }

        }
    }
}
