package be.iminds.ilabt.jfed.experimenter_gui.util;

import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.rspec.model.LoginService;
import be.iminds.ilabt.jfed.rspec.model.ProxyService;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.UserSshKeyInfo;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Singleton;

/**
 * User: twalcari
 * Date: 9/10/2015
 * Time: 10:23
 */
@Singleton
public class ProxyServiceUtil {
    private static final Logger LOG = LoggerFactory.getLogger(ProxyServiceUtil.class);

    private final GeniUserProvider geniUserProvider;

    @Inject
    public ProxyServiceUtil(GeniUserProvider geniUserProvider) {
        this.geniUserProvider = geniUserProvider;
    }


    public JFedConnection.SshProxyInfo findTestbedProxy(@Nonnull RspecNode node, @Nonnull LoginService loginService) {
        assert node.getLoginServices().contains(loginService);

        if (node.getProxyServices().isEmpty())
            return null;

        String loginServiceString = loginServiceToString(loginService);

        for (ProxyService proxyService : node.getProxyServices()) {
            if (loginServiceString.equals(proxyService.getFor())) {
                assert proxyService.getProxy() != null;

                return stringToProxyInfo(proxyService.getProxy(), proxyService.getHostKey());
            }
        }

        return null;
    }

    private static String loginServiceToString(@Nonnull LoginService loginService) {
        return loginService.getUsername() + "@" + loginService.getHostname() + ":" + loginService.getPort();
    }

    private JFedConnection.SshProxyInfo stringToProxyInfo(@Nonnull String proxyString, @Nullable String hostKey) {
        final String originalProxyString = proxyString;

        String username;
        String hostname;
        int port = 22;

        int atPos = proxyString.indexOf("@");

        if (atPos >= 0) {
            username = proxyString.substring(0, atPos);
        } else {
            LOG.warn("No username could be found in proxy-string '{}'", originalProxyString);
            username = null;
        }

        proxyString = proxyString.substring(atPos + 1);
        int semiColonPos = proxyString.lastIndexOf(":");

        if (semiColonPos >= 0) {
            try {
                port = Integer.parseInt(proxyString.substring(semiColonPos + 1));
            } catch (NumberFormatException ex) {
                LOG.error("Could not properly parse port number in proxy-string '{}'", originalProxyString);
                port = 22;
            }

            hostname = proxyString.substring(0, semiColonPos);
        } else {
            LOG.warn("No port number could be found in proxy-string '{}'", originalProxyString);
            hostname = proxyString;
        }

        return new JFedConnection.SshProxyInfo(hostname, port, username,
                new UserSshKeyInfo(geniUserProvider.getLoggedInGeniUser()),
                hostKey, false);

    }

}
