package be.iminds.jfed.gts_highlevel;

import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.util.JFedTrustStore;
import be.iminds.jfed.gts_highlevel.controller.GtsConnectionProvider;
import be.iminds.jfed.gts_highlevel.controller.GtsConnectionProviderImpl;
import be.iminds.jfed.gts_highlevel.controller.GtsLoginConnectionProvider;
import be.iminds.jfed.gts_highlevel.controller.GtsReservationManagerConnectionProvider;
import com.google.inject.AbstractModule;
import com.google.inject.Provides;

import java.net.MalformedURLException;
import java.net.URL;

/**
 * User: twalcari
 * Date: 9/23/2015
 * Time: 11:12
 */
public class GtsModule extends AbstractModule {
    private static final boolean USE_DEBUG_GTS = true;
    private static final String DEBUG_GTS = "http://195.113.161.164:8171";
    private static final String PRODUCTION_GTS = "http://gts.geant.net:8171";

    private final URL loginBaseUrl;
    private final URL reservationManagerBaseUrl;

    public GtsModule() {
        try {
            loginBaseUrl = new URL(USE_DEBUG_GTS ? DEBUG_GTS : PRODUCTION_GTS);
        } catch (MalformedURLException e) {
            throw new RuntimeException("Could not parse loginBaseUrl", e);
        }
        try {
            reservationManagerBaseUrl = new URL(USE_DEBUG_GTS ? DEBUG_GTS : PRODUCTION_GTS + "/rm");
        } catch (MalformedURLException e) {
            throw new RuntimeException("Could not parse reservationManagerBaseUrl", e);
        }
    }

    static {

    }

    @Provides
    @GtsLoginConnectionProvider
    GtsConnectionProvider provideLoginConnectionProvider(JFedTrustStore jFedTrustStore,
                                                         GeniUserProvider geniUserProvider,
                                                         JFedConnectionProvider connectionProvider) {
        return new GtsConnectionProviderImpl(loginBaseUrl, jFedTrustStore, geniUserProvider, connectionProvider);
    }

    @Provides
    @GtsReservationManagerConnectionProvider
    GtsConnectionProvider provideReservationManagerConnectionProvider(JFedTrustStore jFedTrustStore,
                                                                      GeniUserProvider geniUserProvider,
                                                                      JFedConnectionProvider connectionProvider) {
        return new GtsConnectionProviderImpl(reservationManagerBaseUrl, jFedTrustStore, geniUserProvider, connectionProvider);
    }


    @Override
    protected void configure() {
    }
}
