package be.iminds.jfed.gts_highlevel.model;

import be.iminds.jfed.gts_highlevel.util.UnixEpochToZonedDateTimeAdapter;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.List;

/**
 * User: twalcari
 * Date: 9/17/2015
 * Time: 16:47
 */
@XmlRootElement(name="reservation")
public class GtsReservation {
    private Integer id;
    private ZonedDateTime startTime;
    private ZonedDateTime endTime;    //from unix epoch in millis
    private String projectName;
    private List<String> resourcesProviderIds;
    private Status status;
    private List<GtsResource> children;

    public GtsReservation(){}

    public GtsReservation(GtsReservation res){
        this.id = res.id;
        this.startTime = res.startTime;
        this.endTime = res.endTime;
        this.projectName = res.projectName;
        this.resourcesProviderIds = new ArrayList<>(res.resourcesProviderIds);
        this.status = res.status;
        this.children = new ArrayList<>(res.children);
    }

    /**
     * Copy constructor which adapts status to newStatus
     * @param res  GtsReservation to be copied
     * @param newStatus  Status that must be used to override res.status
     */
    public GtsReservation(GtsReservation res, Status newStatus){
        this.id = res.id;
        this.startTime = res.startTime;
        this.endTime = res.endTime;
        this.projectName = res.projectName;
        this.resourcesProviderIds = new ArrayList<>(res.resourcesProviderIds);
        this.status = newStatus;
        this.children = new ArrayList<>(res.children);
    }

    @XmlElement
    public Integer getId() {
        return id;
    }

    void setId(Integer id) {
        this.id = id;
    }

    @XmlJavaTypeAdapter(UnixEpochToZonedDateTimeAdapter.class)
    public ZonedDateTime getStartTime() {
        return startTime;
    }

    void setStartTime(ZonedDateTime startTime) {
        this.startTime = startTime;
    }

    @XmlJavaTypeAdapter(UnixEpochToZonedDateTimeAdapter.class)
    public ZonedDateTime getEndTime() {
        return endTime;
    }

    void setEndTime(ZonedDateTime endTime) {
        this.endTime = endTime;
    }

    @XmlElement
    public String getProjectName() {
        return projectName;
    }

    void setProjectName(String projectName) {
        this.projectName = projectName;
    }

    @XmlElement(name ="resourceProviderId")
    public List<String> getResourcesProviderIds() {
        return resourcesProviderIds;
    }

    void setResourcesProviderIds(List<String> resourcesProviderIds) {
        this.resourcesProviderIds = resourcesProviderIds;
    }

    @XmlElement
    public Status getStatus() {
        return status;
    }

    void setStatus(Status status) {
        this.status = status;
    }

    @XmlElement
    public List<GtsResource> getChildren() {
        return children;
    }

    void setChildren(List<GtsResource> children) {
        this.children = children;
    }

    public enum Status {
        RESERVED,
        ACTIVATING,
        ACTIVE,
        DEACTIVATING,
        DEACTIVED,
        RESERVING, RELEASING
    }


}
