package be.iminds.ilabt.jfed.highlevel.tasks;

import be.iminds.ilabt.jfed.highlevel.controller.JFedHighLevelException;
import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

/**
 * User: twalcari
 * Date: 6/17/2015
 * Time: 15:32
 */
public class DeleteSliceTask extends Task {
    private static final Logger LOG = LoggerFactory.getLogger(DeleteSliceTask.class);

    private final Slice slice;

    private final TaskThread tt;
    private final HighLevelTaskFactory highLevelTaskFactory;

    public DeleteSliceTask(Slice slice, TaskThread tt, HighLevelTaskFactory highLevelTaskFactory) {
        super("Delete All Slivers on slice " + slice.getUrnString());
        this.slice = slice;
        this.tt = tt;
        this.highLevelTaskFactory = highLevelTaskFactory;
    }

    @Override
    public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
        if (slice.isExpired()) {
            throw new JFedHighLevelException("Slice has expired.");
        }
        if (!slice.hasCredential()) {
            throw new JFedHighLevelException("No credentials for slice. Has it expired? slice expiration date = " + slice.getExpirationDate() + " now=" + new Date());
        }

        Set<SfaAuthority> componentManagers = new HashSet<>(slice.getManifestComponentManagers());

        for (SfaAuthority auth : componentManagers) {
            if (auth.isFake())
                continue;
            if (!auth.hasAmUrl()) {
                LOG.warn("Slice has component manager '{}', but that authority has no known AM URL: cannot delete there. Ignoring!", auth.getUrnString());
                continue;
            }

            tt.addTask(highLevelTaskFactory.deleteSliversAtAuthority(slice, auth));
        }
    }

    @Override
    public List<Task> initDependsOn() {
        return Collections.singletonList(highLevelTaskFactory.getSliceCredential(slice));

    }

    @Override
    public List<Task> initAlwaysDependsOn() {
        return Collections.singletonList(highLevelTaskFactory.getAggregatesForSlice(slice));
    }
}
